import argparse
from collections import namedtuple
import copy
import datetime
import itertools as it
import os
from math import modf
from pathlib import Path
import platform
from pprint import pprint
import re
import shutil
import sys
import textwrap
import time

from distributed import Client
import numpy as np
import scipy as sp
import yaml
__version__ = 'Pytorch-Enabled version bundled with `pdb2reaction`'
from pysisyphus.calculators import *
from pysisyphus.config import OUT_DIR_DEFAULT, p_DEFAULT, T_DEFAULT
from pysisyphus.cos import *
from pysisyphus.cos.GrowingChainOfStates import GrowingChainOfStates
from pysisyphus.color import bool_color
from pysisyphus.exceptions import HEIIsFirstOrLastException
from pysisyphus.dynamics import (
    get_mb_velocities_for_geom,
    mdp,
    md,
    get_colvar,
    Gaussian,
)
from pysisyphus.drivers import (
    relaxed_1d_scan,
    run_afir_paths,
    run_opt,
    run_precontr,
    run_perf,
    print_perf_results,
)
from pysisyphus.drivers.barriers import do_endopt_ts_barriers

from pysisyphus.Geometry import Geometry
from pysisyphus.helpers import (
    confirm_input,
    shake_coords,
    print_barrier,
    get_tangent_trj_str,
)
from pysisyphus.helpers_pure import (
    find_closest_sequence,
    merge_sets,
    recursive_update,
    highlight_text,
    approx_float,
    results_to_json,
)
from pysisyphus.intcoords import PrimitiveNotDefinedException
from pysisyphus.intcoords.setup import get_bond_mat
from pysisyphus.init_logging import init_logging
from pysisyphus.intcoords.PrimTypes import PrimTypes, normalize_prim_inputs
from pysisyphus.intcoords.helpers import form_coordinate_union
from pysisyphus.intcoords.setup import get_bond_sets
from pysisyphus.interpolate import interpolate_all
from pysisyphus.irc import *
from pysisyphus.io import save_hessian
from pysisyphus.stocastic import *
from pysisyphus.thermo import can_thermoanalysis
from pysisyphus.trj import get_geoms, dump_geoms, standardize_geoms
from pysisyphus.xyzloader import write_geoms_to_trj
from pysisyphus.yaml_mods import get_loader, UNITS


CALC_DICT = {
    "afir": AFIR,
    "composite": Composite,
    "conical": ConicalIntersection,
    "dftb+": DFTBp,
    "dimer": Dimer,
    "dummy": Dummy,
    "energymin": EnergyMin,
    "ext": ExternalPotential,
    "g09": Gaussian09,
    "g16": Gaussian16,
    "ipiserver": IPIServer,
    "mopac": MOPAC,
    "multi": MultiCalc,
    "obabel": OBabel,
    "oniom": ONIOM,
    "openmolcas": OpenMolcas,
    "orca": ORCA,
    "orca5": ORCA5,
    "psi4": Psi4,
    "pyxtb": PyXTB,
    "remote": Remote,
    "turbomole": Turbomole,
    "xtb": XTB,
    "cfour": CFOUR,
}

try:
    from pysisyphus.calculators.PySCF import PySCF

    CALC_DICT["pyscf"] = PySCF
except ImportError:
    pass

try:
    from pysisyphus.calculators.QCEngine import QCEngine

    CALC_DICT["qcengine"] = QCEngine
except ImportError:
    pass

COS_DICT = {
    "neb": NEB.NEB,
    "aneb": AdaptiveNEB.AdaptiveNEB,
    "feneb": FreeEndNEB.FreeEndNEB,
    "szts": SimpleZTS.SimpleZTS,
    "gs": GrowingString.GrowingString,
    "fs": FreezingString.FreezingString,
}

IRC_DICT = {
    "dvv": DampedVelocityVerlet,
    "euler": Euler,
    "eulerpc": EulerPC,
    "gs": GonzalezSchlegel,
    "imk": IMKMod,
    "lqa": LQA,
    "modekill": ModeKill,
    "rk4": RK4,
}

STOCASTIC_DICT = {
    "frag": FragmentKick,
    "kick": Kick,
}


def parse_args(args):
    parser = argparse.ArgumentParser()

    action_group = parser.add_mutually_exclusive_group(required=True)
    action_group.add_argument(
        "yaml", nargs="?", help="Start pysisyphus with input from a YAML file."
    )
    action_group.add_argument(
        "--clean", action="store_true", help="Ask for confirmation before cleaning."
    )
    action_group.add_argument(
        "--fclean",
        action="store_true",
        help="Force cleaning without prior confirmation.",
    )
    action_group.add_argument(
        "--bibtex",
        action="store_true",
        help="Print bibtex string for pysisyphus paper.",
    )
    action_group.add_argument(
        "-v", "--version", action="store_true", help="Print pysisyphus version."
    )

    run_type_group = parser.add_mutually_exclusive_group(required=False)
    run_type_group.add_argument(
        "--restart",
        action="store_true",
        help="Continue a previously crashed/aborted/... pysisphus run.",
    )
    run_type_group.add_argument(
        "--cp",
        "--copy",
        nargs="+",
        help="Copy .yaml file and corresponding geometries from the 'geom' section "
        "to a new directory. The first argument is interpreted as destination. Any "
        "remaining (optional) arguments are files that are also copied.",
    )

    parser.add_argument(
        "--scheduler", default=None, help="Address of the dask scheduler."
    )
    return parser.parse_args()


def get_calc_closure(base_name, calc_key, calc_kwargs, iter_dict=None, index=None):
    if iter_dict is None:
        iter_dict = dict()

    if index is None:
        index = 0
    # Maps YAML input to actual Calculator-class arguments
    calc_map = {
        "calculator": "calculator",
        "calc": "calculator",  # shortcut for 'calculator'
        # calc1/calc2 are used for ConicalIntersection and EnergyMin.
        "calculator1": "calculator1",
        "calc1": "calculator1",  # shortcut for 'calculator1'
        "calcualtor2": "calculator2",
        "calc2": "calculator2",  # shortcut for 'calculator2'
    }

    def calc_getter(**add_kwargs):
        nonlocal index

        kwargs_copy = copy.deepcopy(calc_kwargs)

        # Some calculators are just wrappers, modifying forces from actual calculators,
        # e.g. AFIR and Dimer. If we find one of the keys in 'calc_map' in 'calc_kwargs'
        # we create the actual calculator and assign it to the corresponding value in
        # 'calc_map'.
        for key, val in calc_map.items():
            if key in kwargs_copy:
                # Use different base_name to distinguish the calculator(s)
                actual_base_name = val
                actual_kwargs = kwargs_copy.pop(key)
                actual_key = actual_kwargs.pop("type")
                # Pass 'index' to arguments, to avoid recreating calculators with
                # the same name.
                actual_calc = get_calc_closure(
                    actual_base_name, actual_key, actual_kwargs, index=index
                )()
                kwargs_copy[val] = actual_calc

        kwargs_copy["base_name"] = base_name
        kwargs_copy.update(add_kwargs)
        kwargs_copy["calc_number"] = index
        for key, iter_ in iter_dict.items():
            kwargs_copy[key] = next(iter_)
        index += 1
        return CALC_DICT[calc_key](**kwargs_copy)

    return calc_getter


def run_tsopt_from_cos(
    cos,
    tsopt_key,
    tsopt_kwargs,
    calc_getter=None,
    ovlp_thresh=0.4,
    coordinate_union="bonds",
    save_hessian=False,
):
    print(highlight_text(f"Running TS-optimization from COS"))

    # Later want a Cartesian HEI tangent, so if not already present we create
    # a Cartesian COS object to obtain the tangent from.
    atoms = cos.images[0].atoms
    if cos.coord_type != "cart":
        cart_images = list()
        for image in cos.images:
            cart_image = Geometry(atoms, image.cart_coords)
            cart_image.energy = image.energy
            cart_images.append(cart_image)
        cart_cos = ChainOfStates.ChainOfStates(cart_images)
    # Just continue using the Cartesian COS object
    else:
        cart_cos = cos

    hei_kind = tsopt_kwargs.pop("hei_kind", "splined")
    # Use plain, unsplined, HEI
    if hei_kind == "plain":
        hei_index = cos.get_hei_index()
        hei_image = cos.images[hei_index]
        # Select the Cartesian tangent from the COS
        cart_hei_tangent = cart_cos.get_tangent(hei_index)
    # Use splined HEI
    elif hei_kind == "splined":
        # The splined HEI tangent is usually very bady for the purpose of
        # selecting an imaginary mode to follow uphill. So we construct a better
        # HEI tangent by mixing the two tangents closest to the splined HEI.
        hei_coords, hei_energy, splined_hei_tangent, hei_index = cos.get_splined_hei()
        hei_image = Geometry(atoms, hei_coords)
        close_to_first = hei_index < 0.5
        close_to_last = hei_index > len(cos.images) - 1.5
        if close_to_first or close_to_last:
            close_to = "first" if close_to_first else "last"
            print(
                f"Splined HEI is too close to the {close_to} image. Aborting TS optimization!"
            )
            raise HEIIsFirstOrLastException
        # The hei_index is a float. We split off the decimal part and mix the two
        # nearest tangents accordingly.
        frac, floor = modf(hei_index)
        # Indices of the two nearest images with integer indices.
        floor = int(floor)
        ceil = floor + 1
        floor_tangent = cart_cos.get_tangent(floor)
        ceil_tangent = cart_cos.get_tangent(ceil)
        print(f"Creating mixed HEI tangent, using tangents at images {(floor, ceil)}.")
        print("Overlap of splined HEI tangent with these tangents:")
        for ind, tang in ((floor, floor_tangent), (ceil, ceil_tangent)):
            print(f"\t{ind:02d}: {splined_hei_tangent.dot(tang):.6f}")
        # When frac is big, e.g. 0.9 the tangent resembles the tangent at 'ceil'
        # so we mix in only (1-frac) == (1-0.9) == 0.1 of the 'floor' tangent.
        cart_hei_tangent = (1 - frac) * floor_tangent + frac * ceil_tangent
        cart_hei_tangent /= np.linalg.norm(cart_hei_tangent)
        # print(f"\t(1-{frac:.4f})*t({floor})+{frac:.4f}*t({ceil}): "
        # f"{splined_hei_tangent.dot(cart_hei_tangent):.6f}"
        # )
    else:
        raise Exception(f"Invalid hei_kind='{hei_kind}'!")

    print(f"Index of {hei_kind} highest energy image (HEI) is {hei_index:.2f}.")
    print()

    # When the COS was optimized in internal coordinates the united primitive
    # indices are already present and we just keep on using them.
    try:
        typed_prims = hei_image.internal.typed_prims
    # If the COS was optimized in Cartesians we have to generated a new
    # set of primitive internals.
    except AttributeError:

        def get_int_geom(geom):
            return Geometry(geom.atoms, geom.cart_coords, coord_type="redund")

        internal_geom1 = get_int_geom(cos.images[0])
        internal_geom2 = get_int_geom(cos.images[-1])
        typed_prims = form_coordinate_union(internal_geom1, internal_geom2)

    coord_kwargs = dict()
    if coordinate_union == "all":
        coord_kwargs["typed_prims"] = typed_prims
        union_msg = "Using full coordinate union for TS guess. Probably a bad idea!"
    elif coordinate_union in ("bonds", "bonds_bends_dihedrals"):
        # Only keep actual bonds ...
        valid_prim_types = (PrimTypes.BOND,)
        # ... and bends and dihedrals, if requested
        if coordinate_union == "bonds_bends_dihedrals":
            valid_prim_types += (PrimTypes.BEND, PrimTypes.PROPER_DIHEDRAL)
        coord_kwargs["define_prims"] = [
            tp for tp in typed_prims if tp[0] in valid_prim_types
        ]
        union_msg = f"Kept primitive types: {valid_prim_types}"
    else:
        union_msg = "No coordinate union."
    print(union_msg)

    ts_geom_kwargs = tsopt_kwargs.pop("geom")
    ts_coord_type = ts_geom_kwargs.pop("type")
    if ts_coord_type != "cart":
        ts_geom_kwargs["coord_kwargs"].update(coord_kwargs)

    ts_geom = Geometry(
        hei_image.atoms,
        hei_image.cart_coords,
        coord_type=ts_coord_type,
        **ts_geom_kwargs,
    )

    # Convert tangent from whatever coordinates to redundant internals.
    # When the HEI was splined the tangent will be in Cartesians.
    if ts_coord_type == "cart":
        ref_tangent = cart_hei_tangent
    elif ts_coord_type in ("redund", "dlc"):
        ref_tangent = ts_geom.internal.B @ cart_hei_tangent
    else:
        raise Exception(f"Invalid coord_type='{ts_coord_type}'!")
    ref_tangent /= np.linalg.norm(ref_tangent)

    # Dump HEI data
    #
    # Cartesian tangent and an animated _trj.xyz file
    cart_hei_fn = "cart_hei_tangent"
    np.savetxt(cart_hei_fn, cart_hei_tangent)
    trj = get_tangent_trj_str(
        ts_geom.atoms, ts_geom.cart_coords, cart_hei_tangent, points=10
    )
    trj_fn = cart_hei_fn + "_trj.xyz"
    with open(trj_fn, "w") as handle:
        handle.write(trj)
    print(f"Wrote animated HEI tangent to {trj_fn}\n")

    # Print HEI information (coords & tangent)
    print(f"{hei_kind.capitalize()} HEI (TS guess)")
    print(ts_geom.as_xyz())
    print()
    dummy = Geometry(atoms, cart_hei_tangent)
    print(f"{hei_kind.capitalize()} Cartesian HEI tangent")
    print(dummy.as_xyz())
    print()

    # Write out HEI information (coords & tangent)
    hei_xyz_fn = f"{hei_kind}_hei.xyz"
    with open(hei_xyz_fn, "w") as handle:
        handle.write(ts_geom.as_xyz())
    print(f"Wrote {hei_kind} HEI coordinates to '{hei_xyz_fn}'")

    ts_calc = calc_getter()

    def wrapped_calc_getter():
        return ts_calc

    ts_geom.set_calculator(ts_calc)
    tsopt_kwargs["prefix"] = "ts"

    if tsopt_key == "dimer":
        # Right now Dimer optimization is rectricted to cartesian
        # rotations and translations, even though translation in
        # internals would be possible.
        ts_geom = Geometry(hei_image.atoms, hei_image.cart_coords)
        dimer_kwargs = tsopt_kwargs.pop("dimer_kwargs", {})
        dimer_kwargs.update(
            {
                "N_raw": cart_hei_tangent,
                "base_name": "dimer",
            }
        )

        def wrapped_calc_getter():
            return Dimer(ts_calc, **dimer_kwargs)

        tsopt_key = "plbfgs"
        # When calling run_opt we pass the Hessian as additional argument,
        # so it is not recalculated unecessary. As no Hessian is available
        # for the dimer method we set it None.
        cart_hessian = None
    # Determine which imaginary mode has the highest overlap with the splined HEI tangent.
    else:
        print(f"Calculating Hessian at {hei_kind} TS guess.")
        # Calculate Hessian
        cart_hessian = ts_geom.cart_hessian
        # Continue Hessian in whatever coordinate system is actually in use
        H = ts_geom.hessian
        eigvals, eigvecs = np.linalg.eigh(H)
        neg_inds = eigvals < -1e-4
        if sum(neg_inds) == 0:
            raise Exception("No negative eigenvalues found at splined HEI. Exiting!")
        eigval_str = np.array2string(eigvals[neg_inds], precision=6)
        print(f"Negative eigenvalues at splined HEI:\n{eigval_str}")
        neg_eigvals = eigvals[neg_inds]
        neg_eigvecs = eigvecs.T[neg_inds]
        ovlps = [np.abs(imag_mode.dot(ref_tangent)) for imag_mode in neg_eigvecs]
        print("Overlaps between HEI tangent and imaginary modes:")
        for i, ov in enumerate(ovlps):
            print(f"\t{i:02d}: {ov:.6f}")
        max_ovlp_ind = np.argmax(ovlps)
        max_ovlp = ovlps[max_ovlp_ind]
        print(
            f"Imaginary mode {max_ovlp_ind} has highest overlap ({max_ovlp:.2%}) "
            "with splined HEI tangent."
        )
        rel_ovlps = np.array(ovlps) / max(ovlps)
        similar_inds = rel_ovlps > 0.80
        # Only 1 big overlap is present
        if (max_ovlp >= ovlp_thresh) and (similar_inds.sum() == 1):
            ovlp_root = np.argmax(ovlps)
        # Multiple big and similar overlaps are present.
        elif (max_ovlp >= ovlp_thresh) and (similar_inds.sum() > 1):
            # Will yield the first occurence of True, which corresponds to a
            # similar overlaps with the most negative eigenvalue.
            ovlp_root = similar_inds.argmax()
            neg_eigval = neg_eigvals[ovlp_root]
            verbose_inds = np.arange(neg_eigvals.size)[similar_inds]
            print(
                f"Overlaps {verbose_inds} are very similar! Falling back to the "
                f"one with the most negative eigenvalue {neg_eigval:.6f} "
                f"(mode {ovlp_root})."
            )
        # Fallback to the most negative eigenvalue when all overlaps are too low.
        else:
            ovlp_root = neg_eigvals.argmin()
            neg_eigval = neg_eigvals[ovlp_root]
            print(
                f"Highest overlap {max_ovlp:.6f} is below the threshold "
                f"of {ovlp_thresh:.6f}.\nFalling back to mode {ovlp_root} with most "
                f"negative eigenvalue {neg_eigval:.6f}."
            )
        root = tsopt_kwargs.get("root", None)
        if root is None:
            # Use mode with highest overlap as initial root
            tsopt_kwargs["root"] = ovlp_root
        else:
            print(
                f"Initial root={root} given, neglecting root {ovlp_root} "
                "determined from overlaps."
            )

    opt_result = run_opt(
        ts_geom,
        calc_getter=wrapped_calc_getter,
        opt_key=tsopt_key,
        opt_kwargs=tsopt_kwargs,
        cart_hessian=cart_hessian,
        title="TS-Optimization",
        copy_final_geom="ts_opt.xyz",
        save_hessian=save_hessian,
    )
    ts_geom = opt_result.geom
    ts_opt = opt_result.opt

    # Restore original calculator for Dimer calculations
    if tsopt_key == "dimer":
        ts_geom.set_calculator(ts_calc)

    print(f"Optimized TS coords:")
    print(ts_geom.as_xyz())
    # Include ts_ prefix
    ts_opt_fn = ts_opt.get_path_for_fn("opt.xyz")
    with open(ts_opt_fn, "w") as handle:
        handle.write(ts_geom.as_xyz())
    print(f"Wrote TS geometry to '{ts_opt_fn}\n'")

    ts_energy = ts_geom.energy
    first_cos_energy = cos.images[0].energy
    last_cos_energy = cos.images[-1].energy
    print_barrier(ts_energy, first_cos_energy, "TS", "first COS image")
    print_barrier(ts_energy, last_cos_energy, "TS", "last COS image")

    print()
    return opt_result


def run_calculations(
    geoms,
    calc_getter,
    scheduler=None,
    assert_track=False,
    run_func=None,
):
    print(highlight_text("Running calculations"))

    func_name = "run_calculation" if run_func is None else run_func

    def par_calc(geom):
        return getattr(geom.calculator, func_name)(geom.atoms, geom.coords)

    for geom in geoms:
        geom.set_calculator(calc_getter())

    if assert_track:
        assert all(
            [geom.calculator.track for geom in geoms]
        ), "'track: True' must be present in calc section."

    if scheduler:
        client = Client(scheduler, pure=False, silence_logs=False)
        results_futures = client.map(par_calc, geoms)
        all_results = client.gather(results_futures)
    else:
        all_results = list()
        i_fmt = "02d"
        for i, geom in enumerate(geoms):
            print(highlight_text(f"Calculation {i:{i_fmt}}", level=1))

            start = time.time()
            print(geom)
            results = getattr(geom.calculator, func_name)(geom.atoms, geom.cart_coords)
            # results dict of MultiCalc will contain keys that can be dumped yet. So
            # we skip the JSON dumping when KeyError is raised.
            try:
                as_json = results_to_json(results)
                calc = geom.calculator
                # Decrease counter, because it will be increased by 1, w.r.t to the
                # calculation.
                json_fn = calc.make_fn("results", counter=calc.calc_counter - 1)
                with open(json_fn, "w") as handle:
                    handle.write(as_json)
            except KeyError:
                print("Skipped JSON dump of calculation results!")

            hess_keys = [
                key
                for key, val in results.items()
                if isinstance(val, dict) and "hessian" in val
            ]
            for hkey in hess_keys:
                hres = results[hkey]
                hfn = f"{hkey}_hessian.h5"
                # save_hessian(
                #     hfn,
                #     geom,
                #     cart_hessian=hres["hessian"],
                #     energy=hres["energy"],
                # )
                # print(f"Dumped hessian to '{hfn}'.")

            all_results.append(results)
            if i < (len(geoms) - 1):
                try:
                    cur_calculator = geom.calculator
                    next_calculator = geoms[i + 1].calculator
                    next_calculator.set_chkfiles(cur_calculator.get_chkfiles())
                    msg = f"Set chkfiles of calculator {i:{i_fmt}} on calculator {i+1:{i_fmt}}"
                except AttributeError:
                    msg = "Calculator does not support set/get_chkfiles!"
                print(msg)
            end = time.time()
            diff = end - start
            print(f"Calculation took {diff:.1f} s.\n")
            sys.stdout.flush()
    print()

    for geom, results in zip(geoms, all_results):
        try:
            geom.set_results(results)
        except KeyError:
            pass

    return geoms, all_results


def run_stocastic(stoc):
    # Fragment
    stoc.run()
    print()

    return stoc


def run_md(geom, calc_getter, md_kwargs):
    print(highlight_text(f"Running Molecular Dynamics"))

    calc = calc_getter()
    geom.set_calculator(calc)

    T_init_vel = md_kwargs.pop("T_init_vel")
    steps = md_kwargs.pop("steps")
    dt = md_kwargs.pop("dt")
    seed = md_kwargs.pop("seed", None)

    _gaussian = md_kwargs.pop("gaussian", {})
    gaussians = list()
    for g_name, g_kwargs in _gaussian.items():
        # Create collective variable
        cv_kwargs = g_kwargs.pop("colvar")
        cv_key = cv_kwargs.pop("type")
        colvar = get_colvar(cv_key, cv_kwargs)

        # Create & append Gaussian
        g_w = g_kwargs.pop("w")
        g_s = g_kwargs.pop("s")
        g_stride = g_kwargs.pop("stride")
        gau = Gaussian(w=g_w, s=g_s, colvar=colvar, dump_name=g_name)
        gaussians.append((g_name, gau, g_stride))

    remove_com_v = md_kwargs.get("remove_com_v")
    v0 = get_mb_velocities_for_geom(
        geom, T_init_vel, seed=seed, remove_com_v=remove_com_v, remove_rot_v=False
    ).flatten()
    md_result = md(geom, v0=v0, steps=steps, dt=dt, gaussians=gaussians, **md_kwargs)

    # from pysisyphus.xyzloader import coords_to_trj
    # trj_fn = "md_trj.xyz"
    # _ = coords_to_trj(
    # trj_fn, geom.atoms, md_result.coords[::md_kwargs["dump_stride"]]
    # )
    print()

    return md_result


def run_scan(geom, calc_getter, scan_kwargs, callback=None):
    print(highlight_text("Relaxed Scan") + "\n")
    assert (
        geom.coord_type != "cart"
    ), "Internal coordinates are required for coordinate scans."

    type_ = scan_kwargs["type"]
    indices = scan_kwargs["indices"]
    steps = scan_kwargs["steps"]
    start = scan_kwargs.get("start", None)
    end = scan_kwargs.get("end", None)
    step_size = scan_kwargs.get("step_size", None)
    symmetric = scan_kwargs.get("symmetric", False)
    # The final prim value is determined either as
    #  start + steps*step_size
    # or
    #  (end - start) / steps .
    #
    # So we always require steps and either end or step_size.
    # bool(a) != bool(b) amounts to an logical XOR.
    assert (steps > 0) and (
        bool(end) != bool(step_size)
    ), "Please specify either 'end' or 'step_size'!"
    if symmetric:
        assert step_size and (
            start is None
        ), "'symmetric: True' requires 'step_size' and 'start == None'!"

    constrain_prims = normalize_prim_inputs(((type_, *indices),))
    constr_prim = constrain_prims[0]

    start_was_none = start is None
    if start_was_none:
        try:
            constr_ind = geom.internal.get_index_of_typed_prim(constr_prim)
        # Recreate geom with appropriate primitive
        except PrimitiveNotDefinedException:
            geom = geom.copy(coord_kwargs={"define_prims": constrain_prims})
            constr_ind = geom.internal.get_index_of_typed_prim(constr_prim)
        # The given indices may not correspond exactly to a typed primitives,
        # as they may be reversed. So we fetch the actual typed primitive.
        constr_prim = geom.internal.typed_prims[constr_ind]
        start = geom.coords[constr_ind]

    if step_size is None:
        step_size = (end - start) / steps
    opt_kwargs = scan_kwargs["opt"].copy()
    opt_key = opt_kwargs.pop("type")

    def wrapper(geom, start, step_size, steps, pref=None):
        return relaxed_1d_scan(
            geom,
            calc_getter,
            [
                constr_prim,
            ],
            start,
            step_size,
            steps,
            opt_key,
            opt_kwargs,
            pref=pref,
            callback=callback,
        )

    if not symmetric:
        scan_geoms, scan_vals, scan_energies = wrapper(geom, start, step_size, steps)
    else:
        # Negative direction
        print(highlight_text("Negative direction", level=1) + "\n")
        minus_geoms, minus_vals, minus_energies = wrapper(
            geom, start, -step_size, steps, pref="minus"
        )
        init_geom = minus_geoms[0].copy()
        # Positive direction. Compared to the negative direction we start at a
        # displaced geometry and reduce the number of steps by 1.
        print(highlight_text("Positive direction", level=1) + "\n")
        plus_start = start + step_size
        # Do one step less, as we already start from the optimized geometry
        plus_steps = steps - 1
        plus_geoms, plus_vals, plus_energies = wrapper(
            init_geom, plus_start, step_size, plus_steps, pref="plus"
        )
        scan_geoms = minus_geoms[::-1] + plus_geoms
        scan_vals = np.concatenate((minus_vals[::-1], plus_vals))
        scan_energies = np.concatenate((minus_energies[::-1], plus_energies))

        trj = "\n".join([geom.as_xyz() for geom in scan_geoms])
        with open("relaxed_scan_trj.xyz", "w") as handle:
            handle.write(trj)
        scan_data = np.stack((scan_vals, scan_energies), axis=1)
        np.savetxt(f"relaxed_scan.dat", scan_data)
    return scan_geoms, scan_vals, scan_energies


def run_preopt(
    first_geom,
    last_geom,
    calc_getter,
    preopt_key,
    preopt_kwargs,
    save_hessian=False,
):
    strict = preopt_kwargs.pop("strict", False)

    geom_kwargs = preopt_kwargs.pop("geom")
    coord_type = geom_kwargs.pop("type")

    def recreate_geom(geom):
        return Geometry(geom.atoms, geom.coords, coord_type=coord_type, **geom_kwargs)

    first_geom = recreate_geom(first_geom)
    last_geom = recreate_geom(last_geom)

    opt_results = list()
    for geom, key in ((first_geom, "first"), (last_geom, "last")):
        # Backup original geometry for RMSD calculation
        org_geom = geom.copy(coord_type="cart")

        prefix = f"{key}_pre"
        opt_kwargs = preopt_kwargs.copy()
        opt_kwargs.update(
            {
                "prefix": prefix,
                "h5_group_name": prefix,
            }
        )
        opt_result = run_opt(
            geom,
            calc_getter,
            preopt_key,
            opt_kwargs,
            title=f"{key} preoptimization",
            save_hessian=save_hessian,
        )
        opt = opt_result.opt
        opt_results.append(opt_result)

        # Continue with next pre-optimization when stopped manually
        if strict and not opt.stopped and not opt.is_converged:
            print(f"Problem in preoptimization of {key}. Exiting!")
            sys.exit()
        print(f"Preoptimization of {key} geometry converged!")
        opt_fn = opt.get_path_for_fn(f"opt.xyz")
        shutil.move(opt.final_fn, opt_fn)
        print(f"Saved final preoptimized structure to '{opt_fn}'.")

        rmsd = org_geom.rmsd(opt_result.geom)
        print(f"RMSD with initial geometry: {rmsd:.6f} au")
        print()

    return opt_results


def run_irc(geom, irc_key, irc_kwargs, calc_getter):
    print(highlight_text(f"Running IRC") + "\n")

    calc = calc_getter()
    calc.base_name = "irc"
    geom.set_calculator(calc, clear=False)

    # Recreate geometry with Cartesian coordinates if needed.
    if geom.coord_type != "cart":
        geom = geom.copy_all(coord_type="cart")

    irc = IRC_DICT[irc_key](geom, **irc_kwargs)
    irc.run()

    return irc


def do_rmsds(xyz, geoms, end_fns, end_geoms, preopt_map=None, similar_thresh=0.25):
    if len(end_fns) != 2 or len(end_geoms) != 2:
        return
    end_fns = [str(fn) for fn in end_fns]
    max_end_len = max(len(s) for s in end_fns)

    if len(geoms) == 1:
        return
    elif len(geoms) > 2:
        geoms = (geoms[0], geoms[-1])
    assert len(geoms) == 2

    if isinstance(xyz, str):
        xyz = (f"{xyz}, first entry", f"{xyz}, last entry")
    elif (not isinstance(xyz, str)) and len(xyz) >= 2:
        xyz = (xyz[0], xyz[-1])
    assert len(xyz) == 2

    # Append original filenames, if supplied
    if preopt_map is not None:
        xyz = [f"{preopt}/{preopt_map[preopt]}" for preopt in xyz]

    max_len = max(len(s) for s in xyz)

    print(highlight_text(f"RMSDs After End Optimizations"))
    print()

    start_cbms = [get_bond_mat(geom) for geom in geoms]
    end_cbms = [get_bond_mat(geom) for geom in end_geoms]
    for i, start_geom in enumerate(geoms):
        fn = xyz[i]
        found_similar = False
        print(f"start geom {i:>2d} ({fn:>{max_len}s})")
        # Condensed bond mat
        start_cbm = start_cbms[i]
        for j, end_geom in enumerate(end_geoms):
            end_fn = end_fns[j]
            # Compare bond matrices
            cbm_match = (start_cbm == end_cbms[j]).all()
            cbm_str = "(bond matrices match)" if cbm_match else ""
            rmsd = start_geom.rmsd(end_geom)
            similar_str = ""
            if rmsd < similar_thresh:
                found_similar = True
                similar_str = " (similar)"
            print(
                f"\tend geom {j:>2d} ({end_fn:>{max_end_len}s}): "
                f"RMSD={rmsd:>8.6f} au{similar_str} " + cbm_str
            )
        if not found_similar:
            print(f"\tRMSDs of end geometries are dissimilar to '{fn}'!")
    print()


def run_endopt(irc, endopt_key, endopt_kwargs, calc_getter, save_hessian=False):
    print(highlight_text(f"Optimizing reaction path ends"))

    # Gather geometries that shall be optimized and appropriate keys.
    to_opt = list()
    if irc.forward:
        coords = irc.all_coords[0]
        to_opt.append((coords, "forward"))
    if irc.backward:
        coords = irc.all_coords[-1]
        to_opt.append((coords, "backward"))
    if irc.downhill:
        coords = irc.all_coords[-1]
        to_opt.append((coords, "downhill"))

    def to_frozensets(sets):
        return [frozenset(_) for _ in sets]

    separate_fragments = endopt_kwargs.pop("fragments", False)
    total = separate_fragments in ("total", False)

    # Convert to array for easy indexing with the fragment lists
    atoms = np.array(irc.atoms)
    fragments_to_opt = list()
    # Expand endpoints into fragments if requested
    for coords, key in to_opt:
        base_name = f"{key}_end"
        c3d = coords.reshape(-1, 3)
        fragments = list()
        fragment_names = list()

        # Detect separate fragments if requested.
        if separate_fragments:
            bond_sets = to_frozensets(get_bond_sets(atoms.tolist(), c3d))
            # Sort atom indices, so the atoms don't become totally scrambled.
            fragments.extend([sorted(frag) for frag in merge_sets(bond_sets)])
            # Disable higher fragment counts. I'm looking forward to the day
            # this ever occurs and someone complains :)
            assert len(fragments) < 10, "Something probably went wrong"
            fragment_names.extend(
                [f"{base_name}_frag{i:02d}" for i, _ in enumerate(fragments)]
            )
            print(f"Found {len(fragments)} fragment(s) at {base_name}")
            for frag_name, frag in zip(fragment_names, fragments):
                print(f"\t{frag_name}: {len(frag)} atoms")

        # Optimize the whole geometries, without splitting them into fragments
        # Skip this optimization if separate fragments are requested and only
        # one fragment is present, which would result in twice the same optimization.
        skip_one_frag = separate_fragments and len(fragments) == 1
        if total and not skip_one_frag:
            # Dummy fragment containing all atom indices.
            fragments.extend(
                [
                    range(len(atoms)),
                ]
            )
            fragment_names.extend(
                [
                    base_name,
                ]
            )
        elif skip_one_frag:
            print(
                f"Only one fragment present for '{key}'. Skipping optimization "
                "of total system."
            )

        fragment_keys = [key] * len(fragments)
        fragment_atoms = [tuple(atoms[list(frag)]) for frag in fragments]
        fragment_coords = [c3d[frag].flatten() for frag in fragments]
        fragments_to_opt.extend(
            list(zip(fragment_keys, fragment_names, fragment_atoms, fragment_coords))
        )
        print()

    to_opt = fragments_to_opt

    geom_kwargs = endopt_kwargs.pop("geom")
    coord_type = geom_kwargs.pop("type")

    results = {k: list() for k in ("forward", "backward", "downhill")}
    for key, name, atoms, coords in to_opt:
        geom = Geometry(
            atoms,
            coords,
            coord_type=coord_type,
            **geom_kwargs,
        )
        initial_fn = f"{name}_initial.xyz"
        with open(initial_fn, "w") as handle:
            handle.write(geom.as_xyz())

        def wrapped_calc_getter():
            calc = calc_getter()
            calc.base_name = name
            return calc

        opt_kwargs = endopt_kwargs.copy()
        opt_kwargs.update(
            {
                "prefix": name,
                "h5_group_name": name,
                "dump": True,
            }
        )
        try:
            opt_result = run_opt(
                geom,
                wrapped_calc_getter,
                endopt_key,
                opt_kwargs,
                title=f"{name} Optimization",
                level=1,
                save_hessian=save_hessian,
            )
        except Exception as err:
            print(f"{err}\nOptimization crashed!")
            continue
        final_fn = opt_result.opt.final_fn
        opt_fn = f"{name}_opt.xyz"
        shutil.move(final_fn, opt_fn)
        print(f"Moved '{final_fn.name}' to '{opt_fn}'.\n")
        results[key].append(opt_result)
    print()
    return results["forward"], results["backward"], results["downhill"]


def run_mdp(geom, calc_getter, mdp_kwargs):
    cwd = Path(".").resolve()
    for i in range(3):
        out_dir = cwd / f"outdir{i:02d}"
        try:
            os.mkdir(out_dir)
        except FileExistsError:
            print(out_dir, "exists")
        add_kwargs = {
            "out_dir": out_dir,
        }
        calc = calc_getter(**add_kwargs)
        geom.set_calculator(calc)
        mdp_result = mdp(geom, **mdp_kwargs)
        break
    return mdp_result


def copy_yaml_and_geometries(run_dict, yaml_fn, dest_and_add_cp, new_yaml_fn=None):
    destination, *copy_also = dest_and_add_cp
    src_path = Path(yaml_fn).resolve().parent
    destination = Path(destination)
    try:
        print(f"Trying to create directory '{destination}' ... ", end="")
        os.mkdir(destination)
        print("done")
    except FileExistsError:
        print("already exists")
    if "geom" in run_dict:
        xyzs = run_dict["geom"]["fn"]
        try:
            xyzs = list(xyzs.values())
        except AttributeError:
            pass
    else:
        xyzs = run_dict["xyz"]
    print("Copying:")
    # Copy geometries
    # When newlines are present we have an inline xyz formatted string
    if not "\n" in xyzs:
        if isinstance(xyzs, str):
            xyzs = [
                xyzs,
            ]
        for xyz in xyzs:
            if xyz.startswith("lib:"):
                continue
            shutil.copy(src_path / xyz, destination)
            print("\t", xyz)
    else:
        print("Found inline xyz formatted string. No files to copy!")
    # Copy additional files
    for src in copy_also:
        try:
            shutil.copy(src_path / src, destination)
            print(f"\t{src}")
        except FileNotFoundError:
            print(f"\tCould not find '{src}'. Skipping!")
    # Update yaml_fn to match destination
    yaml_dest_fn = Path(destination.stem).with_suffix(".yaml")
    shutil.copy(yaml_fn, destination / yaml_dest_fn)
    print("\t", yaml_fn)


def get_defaults(conf_dict, T_default=T_DEFAULT, p_default=p_DEFAULT):
    # Defaults
    dd = {
        "assert": None,
        "afir": None,
        "barrier": None,
        "calc": {
            "type": "dummy",
        },
        "cos": None,
        "endopt": None,
        "geom": None,
        "glob": None,
        "interpol": None,
        "irc": None,
        "md": None,
        "mdp": None,
        "opt": None,
        "perf": None,
        "precontr": None,
        "preopt": None,
        "scan": None,
        "stocastic": None,
        "shake": None,
        "tsopt": None,
    }

    mol_opt_defaults = {
        "dump": True,
        "max_cycles": 150,
        "overachieve_factor": 5,
        "type": "rfo",
        "do_hess": False,
        "T": T_default,
        "p": p_default,
    }
    cos_opt_defaults = {
        "type": "qm",
        "align": True,
        "dump": True,
    }
    if "interpol" in conf_dict:
        dd["interpol"] = {
            "align": True,
        }

    if "cos" in conf_dict:
        dd["cos"] = {
            "type": "neb",
            "fix_first": True,
            "fix_last": True,
        }
        dd["opt"] = cos_opt_defaults.copy()
    # Use a different, more powerful, optimizer when we are not dealing
    # with a COS-optimization.
    elif "opt" in conf_dict:
        dd["opt"] = mol_opt_defaults.copy()
    elif "stocastic" in conf_dict:
        dd["stocastic"] = {
            "type": "frag",
        }

    def get_opt_geom_defaults():
        return {
            "type": "redund",
            "coord_kwargs": {},
        }

    if "tsopt" in conf_dict:
        dd["tsopt"] = mol_opt_defaults.copy()
        dd["tsopt"].update(
            {
                "type": "rsprfo",
                "h5_group_name": "tsopt",
                "prefix": "ts",
            }
        )
        if "cos" in conf_dict:
            dd["tsopt"]["geom"] = get_opt_geom_defaults()

    if "precontr" in conf_dict:
        dd["precontr"] = {
            "prefix": "precontr",
        }

    if "preopt" in conf_dict:
        # We can't just copy dd["opt"] because there will probably be
        # some COS specific optimizer, but we just wan't to optimize the
        # (molecular) endpoints.
        dd["preopt"] = mol_opt_defaults.copy()
        dd["preopt"].update(
            {
                # Optimization specific
                # We are a bit more cautious here
                "max_cycles": 100,
                "thresh": "gau_loose",
                "trust_max": 0.3,
                "geom": get_opt_geom_defaults(),
                # Preopt specific
                "strict": False,
            }
        )
        # dd["preopt"]["geom"]["type"] = "tric"

    if "endopt" in conf_dict:
        dd["endopt"] = mol_opt_defaults.copy()
        dd["endopt"].update(
            {
                "thresh": "gau",
                "fragments": False,
                "geom": get_opt_geom_defaults(),
            }
        )

    if "barriers" in conf_dict:
        dd["barriers"] = {
            "T": T_default,
            "p": p_default,
            "solv_calc": {},
            "do_standard_state_corr": True,
        }

    if "shake" in conf_dict:
        dd["shake"] = {
            "scale": 0.1,
            "seed": None,
        }

    if "irc" in conf_dict:
        dd["irc"] = {
            "type": "eulerpc",
            "rms_grad_thresh": 1e-3,
        }

    if "assert" in conf_dict:
        dd["assert"] = {}

    if "geom" in conf_dict:
        dd["geom"] = {
            "type": "cart",
            "coord_kwargs": {},
        }

    if "mdp" in conf_dict:
        dd["mdp"] = {}

    if "scan" in conf_dict:
        dd["scan"] = {
            "opt": mol_opt_defaults.copy(),
            "symmetric": False,
        }
        dd["scan"]["opt"]["dump"] = False

    if "md" in conf_dict:
        md_T = T_default
        dd["md"] = {
            "T": md_T,
            "T_init_vel": md_T,
            "dt": 0.5,
            "thermostat": "csvr_2",
            "timecon": 50,
            "print_stride": 100,
            "dump_stride": 10,
            "remove_com_v": True,
        }

    if "perf" in conf_dict:
        dd["perf"] = {
            "mems": 2500,
            "repeat": 3,
        }

    if "afir" in conf_dict:
        dd["afir"] = {}

    return dd


def get_last_calc_cycle():
    def keyfunc(path):
        return re.match(r"image_\d+.(\d+).out", str(path))[1]

    cwd = Path(".")
    calc_logs = [str(cl) for cl in cwd.glob("image_*.*.out")]
    calc_logs = sorted(calc_logs, key=keyfunc)
    grouped = it.groupby(calc_logs, key=keyfunc)
    # Find the last completly finished cycle.
    last_length = 0
    last_calc_cycle = 0
    for calc_cycle, group in grouped:
        cycle_length = len(list(group))
        if cycle_length < last_length:
            # When this is True we have a cycle that has less
            # items than last one, that is an unfinished cycle.
            break
        last_length = cycle_length
        last_calc_cycle = int(calc_cycle)
    if last_calc_cycle == 0:
        print("Can't find any old calculator logs.")
    print(f"Last calculation counter is {last_calc_cycle}.")
    return last_calc_cycle


VALID_KEYS = {
    "assert",
    "afir",
    "barriers",
    "calc",
    "cos",
    "endopt",
    "geom",
    "interpol",
    "irc",
    "md",
    "mdp",
    "opt",
    "perf",
    "precontr",
    "preopt",
    "scan",
    "shake",
    "stocastic",
    "tsopt",
}


def setup_run_dict(run_dict):
    org_dict = run_dict.copy()

    # Load defaults to have a sane baseline
    run_dict = get_defaults(run_dict)
    # Update nested entries that are dicts by themselves
    # Take care to insert a , after the string!
    key_set = set(org_dict.keys())
    assert (
        key_set <= VALID_KEYS
    ), f"Found invalid keys in YAML input: {key_set - VALID_KEYS}"
    for key in key_set & VALID_KEYS:
        try:
            # Recursive update, because there may be nested dicts
            recursive_update(run_dict[key], org_dict[key])
        except TypeError:
            print(f"Using default values for '{key}' section.")
    return run_dict


RunResult = namedtuple(
    "RunResult",
    (
        "preopt_first_geom preopt_last_geom "
        "cos cos_opt "
        "ts_geom ts_opt "
        "end_geoms irc irc_geom "
        "mdp_result "
        "opt_geom opt "
        "calced_geoms calced_results "
        "stocastic calc_getter "
        "scan_geoms scan_vals scan_energies "
        "perf_results "
    ),
)


def main(run_dict, restart=False, yaml_dir="./", scheduler=None):

    # Dump run_dict
    run_dict_copy = run_dict.copy()
    run_dict_copy["version"] = __version__
    with open("RUN.yaml", "w") as handle:
        yaml.dump(run_dict_copy, handle)

    if run_dict["interpol"]:
        interpol_key = run_dict["interpol"].pop("type")
        interpol_kwargs = run_dict["interpol"]
    # Preoptimization prior to COS optimization
    if run_dict["preopt"]:
        preopt_key = run_dict["preopt"].pop("type")
        preopt_kwargs = run_dict["preopt"]
    # Optimization of fragments after IRC integration
    if run_dict["endopt"]:
        endopt_key = run_dict["endopt"].pop("type")
        endopt_kwargs = run_dict["endopt"]
    if run_dict["opt"]:
        opt_key = run_dict["opt"].pop("type")
        opt_kwargs = run_dict["opt"]
    if run_dict["cos"]:
        cos_key = run_dict["cos"].pop("type")
        cos_kwargs = run_dict["cos"]
        cos_kwargs["scheduler"] = scheduler
    if run_dict["stocastic"]:
        stoc_key = run_dict["stocastic"].pop("type")
        stoc_kwargs = run_dict["stocastic"]
    if run_dict["tsopt"]:
        tsopt_key = run_dict["tsopt"].pop("type")
        tsopt_kwargs = run_dict["tsopt"]
    if run_dict["irc"]:
        irc_key = run_dict["irc"].pop("type")
        irc_kwargs = run_dict["irc"]
    if run_dict["afir"]:
        afir_key = run_dict["afir"].pop("type")
        afir_kwargs = run_dict["afir"]

    # Handle geometry input. This section must always be present.
    geom_kwargs = run_dict["geom"]
    xyz = geom_kwargs.pop("fn")
    coord_type = geom_kwargs.pop("type")
    union = geom_kwargs.pop("union", None)

    ####################
    # CALCULATOR SETUP #
    ####################

    # Prepare calculator
    calc_key = run_dict["calc"].pop("type")
    save_hessian = run_dict["calc"].pop("save_hessian", False)
    calc_kwargs = run_dict["calc"]
    calc_run_func = calc_kwargs.pop("run_func", None)
    calc_kwargs["out_dir"] = calc_kwargs.get("out_dir", yaml_dir / OUT_DIR_DEFAULT)
    calc_base_name = calc_kwargs.get("base_name", "calculator")
    if calc_key in ("oniom", "ext"):
        geoms = get_geoms(xyz, quiet=True)
        iter_dict = {
            "geom": iter(geoms),
        }
    elif calc_key == "multi":
        geoms = get_geoms(xyz, quiet=True)
        iter_dict = {
            "base_name": iter([geom.name for geom in geoms]),
        }
    else:
        iter_dict = None
    calc_getter = get_calc_closure(
        calc_base_name, calc_key, calc_kwargs, iter_dict=iter_dict
    )
    # Create second function that returns a wrapped calculator. This may be
    # useful if we later want to drop the wrapper and use the actual calculator.
    if "calc" in calc_kwargs:
        act_calc_kwargs = calc_kwargs["calc"].copy()
        act_calc_key = act_calc_kwargs.pop("type")
        act_calc_getter = get_calc_closure(
            "act_calculator", act_calc_key, act_calc_kwargs
        )
    try:
        solv_calc_kwargs = run_dict["barriers"].pop("solv_calc")
        solv_calc_key = solv_calc_kwargs.pop("type")
        solv_calc_getter = get_calc_closure(
            "solv_calculator", solv_calc_key, solv_calc_kwargs
        )
    except KeyError:
        solv_calc_getter = None

    ##################
    # GEOMETRY SETUP #
    ##################

    # Initial loading of geometries from file(s)
    geoms = get_geoms(xyz, coord_type="cart")

    # ------------------------+
    #   Preconditioning of   |
    # Translation & Rotation |
    # ------------------------+

    if run_dict["precontr"]:
        ptr_geom0, ptr_geom_m1 = run_precontr(
            geoms[0], geoms[1], prefix=run_dict["precontr"]["prefix"]
        )
        geoms[0] = ptr_geom0
        geoms[-1] = ptr_geom_m1

    # -----------------------+
    # Preoptimization of    |
    # first & last image(s) |
    # -----------------------+

    # Preoptimization only makes sense with a subsequent COS run.
    if run_dict["preopt"] and (run_dict["cos"] or run_dict["afir"]):
        assert len(geoms) > 1
        # Preopt should be expanded to support > 2 fragments with AFIR
        if run_dict["afir"] and len(geoms) > 2:
            raise Exception("Currently, only the first & last geometry are optimized!")
        first_opt_result, last_opt_result = run_preopt(
            geoms[0],
            geoms[-1],
            calc_getter,
            preopt_key,
            preopt_kwargs,
            save_hessian=save_hessian,
        )
        # Update with (pre)optimized geometries. 'preopt_first_geom'/'preopt_last_geom'
        # are assigned here so they can later be assigned to 'run_result':
        geoms[0] = preopt_first_geom = first_opt_result.geom.copy(coord_type=coord_type)
        geoms[-1] = preopt_last_geom = last_opt_result.geom.copy(coord_type=coord_type)

    if run_dict["interpol"]:
        # Interpolate. Will return the original geometries for between = 0
        geoms = interpolate_all(geoms, kind=interpol_key, **interpol_kwargs)
        dump_geoms(geoms, "interpolated")

    # Recreate geometries with desired coordinate type and keyword arguments
    geoms = standardize_geoms(geoms, coord_type, geom_kwargs, union=union)

    # Create COS objects and supply a function that yields new Calculators,
    # as needed for growing COS classes, where images are added over time.
    if run_dict["cos"]:
        cos_cls = COS_DICT[cos_key]
        if issubclass(cos_cls, GrowingChainOfStates) or isinstance(
            cos_cls, type(FreezingString)
        ):
            cos_kwargs["calc_getter"] = get_calc_closure("image", calc_key, calc_kwargs)
        geom = COS_DICT[cos_key](geoms, **cos_kwargs)
    elif len(geoms) == 1:
        geom = geoms[0]

    if run_dict["stocastic"]:
        stoc_kwargs["calc_kwargs"] = calc_kwargs
        stocastic = STOCASTIC_DICT[stoc_key](geom, **stoc_kwargs)
        stocastic = run_stocastic(stocastic)
    elif run_dict["md"]:
        md_kwargs = run_dict["md"].copy()
        run_md(geom, calc_getter, md_kwargs)
    elif run_dict["scan"]:
        scan_kwargs = run_dict["scan"]
        scan_geoms, scan_vals, scan_energies = run_scan(geom, calc_getter, scan_kwargs)
    elif run_dict["perf"]:
        perf_results = run_perf(geom, calc_getter, **run_dict["perf"])
        print_perf_results(perf_results)
    elif run_dict["afir"]:
        ts_guesses, afir_paths = run_afir_paths(
            afir_key, geoms, calc_getter, **afir_kwargs,
        )
    # This case will handle most pysisyphus runs. A full run encompasses
    # the following steps:
    #
    #    (0. Preoptimization, already handled)
    #     1. (COS)-Optimization
    #     2. TS-Optimization by TSHessianOptimizer or Dimer method
    #     3. IRC integration
    #     4. Optimization of IRC endpoints
    #
    # Everything can be chained. All functions operate on the 'geom' object,
    # which is propagated along through all functions calls.
    #
    # All keys are present in 'run_dict', but most of the corresponding values will
    # be set to zero.
    elif any(
        [run_dict[key] is not None for key in ("opt", "tsopt", "irc", "mdp", "endopt")]
    ):

        #######
        # OPT #
        #######

        if run_dict["opt"]:
            if run_dict["shake"]:
                print(highlight_text("Shake coordinates"))
                shaked_coords = shake_coords(geom.coords, **run_dict["shake"])
                geom.coords = shaked_coords
                print(f"Shaken coordinates:\n{geom.as_xyz()}")
            opt_result = run_opt(
                geom,
                calc_getter,
                opt_key,
                opt_kwargs,
                print_thermo=True,
                save_hessian=save_hessian,
            )
            opt_geom = opt_result.geom
            opt = opt_result.opt
            # Keep a backup of the optimized geometry
            if isinstance(opt_geom, ChainOfStates.ChainOfStates):
                # Set some variables that are later collected into RunResult
                cos = opt_geom
                cos_opt = opt
                # copy() is not present for ChainOfState objects, so we just keep
                # using the COS object with a different name.
                geom = opt_geom
            else:
                geom = opt_geom.copy()

        #########
        # TSOPT #
        #########

        abort = False
        if run_dict["tsopt"]:
            # Use a separate implementation for TS-Optimizations started from
            # COS-optimizations.
            try:
                if isinstance(geom, ChainOfStates.ChainOfStates):
                    ts_calc_getter = get_calc_closure(tsopt_key, calc_key, calc_kwargs)
                    ts_opt_result = run_tsopt_from_cos(
                        geom,
                        tsopt_key,
                        tsopt_kwargs,
                        ts_calc_getter,
                        save_hessian=save_hessian,
                    )
                else:
                    ts_opt_result = run_opt(
                        geom,
                        calc_getter,
                        tsopt_key,
                        tsopt_kwargs,
                        title="TS-Optimization",
                        copy_final_geom="ts_opt.xyz",
                        save_hessian=save_hessian,
                    )
                ts_geom = ts_opt_result.geom
                ts_opt = ts_opt_result.opt
                geom = ts_geom.copy_all()
            except HEIIsFirstOrLastException:
                abort = True

        #######
        # IRC #
        #######

        ran_irc = False
        if (not abort) and run_dict["irc"]:
            # After a Dimer run we continue with the actual calculator
            # and not the Dimer calculator.
            if calc_key == "dimer":
                calc_getter = act_calc_getter
            irc_geom = geom.copy()
            irc = run_irc(geom, irc_key, irc_kwargs, calc_getter)
            # IRC geom won't have a calculator, so we set the appropriate values here.
            irc_geom.energy = irc.ts_energy
            # irc_geom.cart_hessian = irc.init_hessian
            ran_irc = True

        ##########
        # ENDOPT #
        ##########

        # Run 'endopt' when a previous IRC calculation was done
        # if ran_irc and run_dict["endopt"]:
        if run_dict["endopt"]:
            if not ran_irc:

                _, irc_geom, _ = geoms  # IRC geom should correspond to the TS

                class DummyIRC:
                    def __init__(self, geoms):
                        first, _, last = geoms
                        self.atoms = copy.copy(first.atoms)
                        self.forward = True
                        self.backward = True
                        self.downhill = False
                        self.all_coords = [
                            geom.cart_coords.copy() for geom in (first, last)
                        ]
                        self.hessian_init = "dummy"

                irc = DummyIRC(geoms)

            do_thermo = run_dict["endopt"].get("do_hess", False) and can_thermoanalysis
            T = run_dict["endopt"]["T"]
            p = run_dict["endopt"]["p"]
            # Order is forward, backward, downhill
            endopt_results = run_endopt(
                irc,
                endopt_key,
                endopt_kwargs,
                calc_getter,
                save_hessian=save_hessian,
            )

            # Determine "left" and "right" geoms
            # Only downhill
            if irc.downhill:
                left_results = endopt_results[2]
            # Only backward
            elif irc.backward and not irc.forward:
                left_results = endopt_results[1]
            # Forward and backward run
            else:
                left_results = endopt_results[0]

            left_geoms = [result.geom for result in left_results]
            left_fns = [result.fn for result in left_results]
            # Use 'backward' results; might be empty.
            right_geoms = [result.geom for result in endopt_results[1]]
            right_fns = [result.fn for result in endopt_results[1]]

            end_geoms = left_geoms + right_geoms
            if run_dict["cos"]:
                end_fns = left_fns + right_fns
                do_rmsds(xyz, geoms, end_fns, end_geoms)

            # Try to compute barriers.
            barrier_kwargs = {
                "do_thermo": do_thermo,
                "T": T,
                "p": p,
                "calc_getter": calc_getter,
                "solv_calc_getter": solv_calc_getter,
            }
            barrier_kwargs_ = run_dict.get("barriers", {})
            barrier_kwargs.update(barrier_kwargs_)
            do_endopt_ts_barriers(
                irc_geom,
                left_geoms,
                right_geoms,
                left_fns=left_fns,
                right_fns=right_fns,
                **barrier_kwargs,
            )

            # Dump TS and endopt geoms to _trj.xyz. But only when we did not optimize
            # separate fragments.
            if len(left_geoms) == 1 and len(right_geoms) in (0, 1):
                trj_fn = "left_ts_right_geoms_trj.xyz"
                write_geoms_to_trj(
                    list(it.chain(left_geoms, [irc_geom], right_geoms)),
                    trj_fn,
                    comments=list(it.chain(left_fns, ["IRC start"], right_fns)),
                )
                print(f"Wrote optimized end-geometries and TS to '{trj_fn}'")

        if run_dict["mdp"]:
            mdp_kwargs = run_dict["mdp"]
            mdp_result = run_mdp(geom, calc_getter, mdp_kwargs)
    # Fallback when no specific job type was specified
    else:
        calced_geoms, calced_results = run_calculations(
            geoms, calc_getter, scheduler, run_func=calc_run_func
        )

    # We can't use locals() in the dict comprehension, as it runs in its own
    # local scope.
    locals_ = locals()
    results = {key: locals_.get(key, None) for key in RunResult._fields}
    run_result = RunResult(**results)
    return run_result


def check_asserts(results, run_dict):
    print(highlight_text(f"Asserting results"))

    assert_ = run_dict["assert"]
    keys = list(assert_.keys())
    objs_attrs = [key.split(".") for key in keys]
    ref_vals = [assert_[k] for k in keys]
    matches = list()
    for i, ((obj, attr), ref_val) in enumerate(zip(objs_attrs, ref_vals)):
        cur_val = getattr(getattr(results, obj), attr)
        matched = approx_float(cur_val, ref_val)
        print(f"{i:02d}: {obj}.{attr}")
        print(f"\tReference: {ref_val}")
        print(f"\t  Current: {cur_val}")
        print(f"\t  Matches: {bool_color(matched)}")
        matches.append(matched)

    assert all(matches)
    print()


def do_clean(force=False):
    """Deletes files from previous runs in the cwd.
    A similar function could be used to store everything ..."""
    cwd = Path(".").resolve()
    rm_globs = (
        "cycle*_trj.xyz",
        "interpolated_trj.xyz",
        "interpolated.image*.xyz",
        "calculator.log",
        "optimizer.log",
        "tsoptimizer.log",
        "wfoverlap.log",
        "host_*.calculator.log",
        "host_*.wfoverlap.log",
        "wfo_*.out" "optimization_trj.xyz",
        "cos.log",
        "*.gradient",
        "optimizer_results.yaml",
        # ORCA specific
        "*.orca.gbw",
        "*.orca.cis",
        "*.orca.engrad",
        "*.orca.hessian",
        "*.orca.inp",
        "*.orca.hess",
        "*.orca.molden",
        # OpenMOLCAS specific
        "calculator*.out",
        "calculator*.JobIph",
        "calculator*.RasOrb",
        "*rasscf.molden",
        # Turbomole specific
        "calculator_*.control",
        "calculator_*.coord",
        "calculator_*.mos",
        "calculator_*.ciss_a",
        "calculator*.sing_a",
        "*wavefunction.molden",
        "*input.xyz",
        "*.coord",
        # PySCF specific
        "calculator*.chkfile",
        "*.pyscf.out",
        "*.chkfile",
        # WFOverlap specific
        "wfo_*.*.out",
        # XTB specific
        "image*.grad",
        "calculator*.grad",
        "calculator*.xcontrol",
        "calculator*.charges",
        "image_*",
        "splined_ts_guess.xyz",
        "splined_hei_tangent",
        "cart_hei_tangent_trj.xyz",
        "dimer_ts.xyz",
        "dimer_pickle",
        "interpolated.geom_*.xyz",
        # Wavefunction overlap
        "wfo_*",
        "image*.molden",
        "jmol.spt",
        "overlap_data.h5",
        "*_CDD.png",
        "*_CDD.cub",
        "internal_coords.log",
        "hei_tangent",
        "optimization_trj.xyz",
        "splined_hei.xyz",
        "ts_opt.xyz",
        "final_geometry.xyz",
        "calculated_init_hessian",
        "cur_out",
        # HDF5 files
        "optimization.h5",
        "afir.h5",
        # Optimization files
        "*_optimization_trj.xyz",
        # Preopt files
        "first_*",
        "last_*",
        # TSOpt
        "rsirfo*",
        # IRC files
        "irc_*",
        "irc.log",
        "finished_*",
        # IRC/Endopt files
        "backward_*",
        "forward_*",
        # Misc
        "*imaginary_mode_*_trj.xyz",
        "cart_hei_tangent",
        "ts_calculated_init_cart_hessian",
        "calculated_final_cart_hessian",
        "*final_geometry.xyz",
        "*final_geometries_trj.xyz",
        "current_geometry.xyz",
        "*current_geometries_trj.xyz",
        "hess_calc_cyc*.h5",
        "ts_hess_calc_cyc*.h5",
        "hess_init_irc.h5",
        "final_hessian.h5",
        "ts_current_geometry.xyz",
        "dimer_*",
        "plain_hei_tangent",
        "plain_hei.xyz",
        "hess_calc_irc*.h5",
        "rebuilt_primitives.xyz",
        "RUN.yaml",
        "middle_for_preopt_trj.xyz",
        "relaxed_scan_trj.xyz",
        "too_similar_trj.xyz",
        # MDP
        "mdp_ee_ascent_trj.xyz",
        "mdp_ee_fin_*_trj.xyz",
        "mdp_ee_init_*_trj.xyz",
        "aligned.geom*xyz",
        "cos_hei_trj.xyz",
        # Dimer
        "calculator_*.N",
        "calculator_*.N_trj.xyz",
        "dimer.log",
        "*.gfnff_topo",
        # DFTB+
        "*.detailed.out",
        "*.geometry.gen",
        "*.dftb_in.hsd",
        "*.EXC.DAT",
        "*.XplusY.DAT",
        "*.dftb.out",
        "rsprfo_*",
        "reparametrized_trj.xyz",
        "end_geoms_and_ts_trj.xyz",
        "left_ts_right_geoms_trj.xyz",
        "ts_final_hessian.h5",
        "third_deriv.h5",
        "*.ao_ovlp_rec",
        # MOPAC
        "*.mopac.aux",
        "*.mopac.arc",
        "*.mopac.mop",
        "*.mopac.out",
    )
    to_rm_paths = list()
    for glob in rm_globs:
        to_rm_paths.extend(list(cwd.glob(glob)))
    to_rm_strs = [str(p) for p in to_rm_paths]
    for s in to_rm_strs:
        print(s)

    def delete():
        for p in to_rm_paths:
            try:
                os.remove(p)
                print(f"Deleted {p}")
            except FileNotFoundError:
                pass
        try:
            os.unlink("cur_out")
        except FileNotFoundError:
            pass

    if force:
        delete()
        return
    # If we dont force the cleaning ask for confirmation first
    elif to_rm_paths and confirm_input("Delete these files?"):
        delete()
    else:
        print("No files found for removal.")


def print_header():
    """Generated from https://asciiartgen.now.sh/?s=pysisyphus&style=colossal"""
    normal_logo = """                           d8b                            888
                           Y8P                            888
                                                          888
88888b.  888  888 .d8888b  888 .d8888b  888  888 88888b.  88888b.  888  888 .d8888b
888 "88b 888  888 88K      888 88K      888  888 888 "88b 888 "88b 888  888 88K
888  888 888  888 "Y8888b. 888 "Y8888b. 888  888 888  888 888  888 888  888 "Y8888b.
888 d88P Y88b 888      X88 888      X88 Y88b 888 888 d88P 888  888 Y88b 888      X88
88888P"   "Y88888  88888P' 888  88888P'  "Y88888 88888P"  888  888  "Y88888  88888P'
888           888                            888 888
888      Y8b d88P                       Y8b d88P 888
888       "Y88P"                         "Y88P"  888                            """

    xmas_logo = r"""
                                   \|/
                           d8b    --o--                   888              X
                  x        Y8P     /|\                    888
                                                          888
88888b.  888  888 .d8888b  888 .d8888b  888  888 88888b.  88888b.  888  888 .d8888b
888 "88b 888  888 88K      888 88K      888  888 888 "88b 888 "88b 888  888 88K
888  888 888  888 "Y8888b. 888 "Y8888b. 888  888 888  888 888  888 888  888 "Y8888b.
888 d88P Y88b 888      X88 888      X88 Y88b 888 888 d88P 888  888 Y88b 888      X88
88888P"   "Y88888  88888P' 888  88888P'  "Y88888 88888P"  888  888  "Y88888  88888P'
888  |        888   |  |    |     |          888 888       |    |              |
888  O   Y8b d88P   o  |    X    / \    Y8b d88P 888       x   / \             O
888       "Y88P"       O         \_/     "Y88P"  888           \_/              
            |                         x           |                      \|/
   X        |                        xox          O                     --X--
            O                         x                                  /|\
"""
    now = datetime.datetime.now()
    today = now.date()
    xmas = (today.month == 12) and (today.day <= 24)
    logo = xmas_logo if xmas else normal_logo
    version = f"Version {__version__}"
    try:
        commit = re.compile(r"g(\w+)\.").search(version).group(1)
        commit_line = f"Git commit {commit}\n"
    # Raised when regex search failed
    except AttributeError:
        commit_line = ""
    vi = sys.version_info
    sv = f"{vi.major}.{vi.minor}.{vi.micro}"  # Python
    npv = np.__version__  # Numpy
    spv = sp.__version__  # SciPy
    cwd = Path(".").resolve()
    print(
        f"{logo}\n\n{version} (Python {sv}, NumPy {npv}, SciPy {spv})\n"
        f"{commit_line}"
        f"Executed at {now.strftime('%c')} on '{platform.node()}'\n"
        f"Platform: {platform.platform()}\n"
        f"Interpreter: {sys.executable}\n"
        f"Current working directory: {cwd}\n"
    )


def print_bibtex():
    bibtex = textwrap.dedent(
        """@article{Steinmetzer2020,
        doi = {10.1002/qua.26390},
        url = {https://doi.org/10.1002/qua.26390},
        year = {2020},
        month = aug,
        publisher = {Wiley},
        author = {Johannes Steinmetzer and Stephan Kupfer and Stefanie Gräfe},
        title = {pysisyphus: Exploring potential energy surfaces in ground and excited states},
        journal = {International Journal of Quantum Chemistry}
    }"""
    )
    print(bibtex)


def run_from_dict(
    run_dict,
    cwd=None,
    set_defaults=True,
    yaml_fn=None,
    cp=None,
    scheduler=None,
    clean=False,
    fclean=False,
    version=False,
    restart=False,
):
    if cwd is None:
        cwd = Path(".")

    print_header()

    # Citation
    citation = (
        "If pysisyphus benefitted your research please cite:\n\n"
        "\thttps://doi.org/10.1002/qua.26390\n\nGood luck!\n"
    )
    print(citation)

    init_logging(cwd, scheduler)
    # Load defaults etc.
    if set_defaults:
        run_dict = setup_run_dict(run_dict)
    sys.stdout.flush()

    if cp:
        copy_yaml_and_geometries(run_dict, yaml_fn, cp)
        return
    elif clean:
        do_clean()
        return
    elif fclean:
        do_clean(force=True)
        return
    # Return after header was printed
    elif version:
        return

    run_dict_without_none = {k: v for k, v in run_dict.items() if v is not None}
    pprint(run_dict_without_none, compact=True, width=200)
    print()
    sys.stdout.flush()

    run_result = main(run_dict, restart, cwd, scheduler)

    if run_dict["assert"] is not None:
        print()
        check_asserts(run_result, run_dict)

    return run_result


def load_run_dict(yaml_fn):
    with open(yaml_fn) as handle:
        yaml_str = handle.read()
    try:
        loader = get_loader()
        try:
            run_dict = yaml.load(yaml_str, Loader=loader)
        except yaml.constructor.ConstructorError as err:
            mobj = re.compile(r"for the tag '\!(\w+)'").search(err.problem)
            if mobj:
                err_unit = mobj.group(1)
                best_match, _ = find_closest_sequence(err_unit, UNITS)
                print(
                    f"Unknown unit!\nKnown units are\n'{UNITS}'.\n"
                    f"Did you mean '{best_match}', instead of '{err_unit}'?\n"
                )
            raise err
        assert type(run_dict) == type(dict())
    except (AssertionError, yaml.parser.ParserError) as err:
        print(err)
        if not (yaml_fn.lower().endswith(".yaml")):
            print("Are you sure that you supplied a YAML file?")
        sys.exit(1)
    return run_dict


def run():
    start_time = datetime.datetime.now()
    args = parse_args(sys.argv[1:])

    # Defaults
    run_dict = {}
    yaml_dir = Path(".")

    if args.yaml:
        run_dict = load_run_dict(args.yaml)
        yaml_dir = Path(os.path.abspath(args.yaml)).parent
    elif args.bibtex:
        print_bibtex()
        return

    run_kwargs = {
        "cwd": yaml_dir,
        "set_defaults": True,
        "yaml_fn": args.yaml,
        "cp": args.cp,
        "scheduler": args.scheduler,
        "clean": args.clean,
        "fclean": args.fclean,
        "version": args.version,
        "restart": args.restart,
    }
    run_result = run_from_dict(run_dict, **run_kwargs)

    end_time = datetime.datetime.now()
    duration = end_time - start_time
    # Only keep hh:mm:ss
    duration_hms = str(duration).split(".")[0]
    print(f"pysisyphus run took {duration_hms} h.")

    return 0


if __name__ == "__main__":
    run()



# [1] https://doi.org/10.1002/jcc.27169
#     Nudged elastic stiffness band method: A method to solve kinks problems
#     of reaction paths
#     Mitsuta, Asada

# Adapted from the original implementation provided under MIT
# license by Yuki Mitsuta. Modifications by Johannes Steinmetzer.
#
# Main modifications were made in 'get_stiff_stress'; the other two
# functions are mostly original, disregarding some renaming.
#
# Original code is found at https://github.com/YukiMitsuta/Fstiffness


from collections.abc import Sequence
from typing import Union

import numpy as np


def get_stiff_stress(
    bandwidth: float,
    kappa: Union[float, Sequence],
    image_coords: np.ndarray,
    tangents: np.ndarray,
    straight_thresh: float = 1e-8,
) -> np.ndarray:
    """Function to calculate the stress of stiffness

    See [1] for a discussion.

    Parameters
    ----------
    bandwidth
        The band width of stiffness term in Bohr. The larger this parameter is,
        the higher the stress of stiffness.
    kappa
        The NEB spring constant (float or sequence of float). When a float is given
        the same spring constants is used for all images.
    image_coords
        2d array of image coordinates of shape (nimages, ncoords) with nimages
        being the number of images and ncoords the number of coordinates of
        each image the instance of each coordinates are numpy.array
    tangents
        2d array with the same shape as 'image_coords', containing the tangents
        defined at each image.

    Returns
    -------
    Fstlist
        Array containing the stress of stiffness with the same shape as 'image_coords'.
    """
    assert bandwidth > 0.0
    assert image_coords.shape == tangents.shape

    nimages, ncoords = image_coords.shape
    Fstlist = np.zeros((nimages, ncoords))

    # Determine if all images are in a straight line by calculating the dot products
    # of all tangents with the normalized coordinate difference vector between the first
    # and last image.
    start_end_tangent = image_coords[-1] - image_coords[0]
    start_end_tangent /= np.linalg.norm(start_end_tangent)
    ovlps = np.einsum("u,vu->v", start_end_tangent, tangents)
    if (np.abs(ovlps - 1.0) <= straight_thresh).all():
        return Fstlist

    tangents_perp, _ = calcEholo_vert(image_coords, tangents)

    offset = 0.5 * tangents_perp * bandwidth
    # Virtual image coordinates. Eq. (14) and (15) in [1].
    plus_coords = image_coords.copy() + offset
    minus_coords = image_coords.copy() - offset

    # Construct list of kappas, when only one kappa is given.
    # For N images, pysisyphus uses N-1 spring constants
    try:
        len(kappa)
    except TypeError:
        kappa = [kappa] * (nimages - 1)

    assert len(kappa) == (nimages - 1)

    # Terminal images don't feel stiffness force.
    for k in range(1, nimages - 1):
        Fstlist[k] = Fstiffness_k(
            k,
            kappa[k - 1],
            kappa[k],
            minus_coords,
            plus_coords,
            tangents_perp,
        )
    return Fstlist


def Fstiffness_k(
    k: int,
    kappa0: float,
    kappa1: float,
    innerimagelist: np.ndarray,
    outerimagelist: np.ndarray,
    Eholo_vertical: np.ndarray,
):
    """Stress of stiffness for image k."""

    v0inner = innerimagelist[k] - innerimagelist[k + 1]
    tdelta0inner = np.linalg.norm(v0inner)
    v0outer = outerimagelist[k] - outerimagelist[k + 1]
    tdelta0outer = np.linalg.norm(v0outer)
    tdeltadelta = tdelta0inner - tdelta0outer
    Vvert = -Eholo_vertical[k] * tdeltadelta * kappa0 * 0.5

    v1inner = innerimagelist[k] - innerimagelist[k - 1]
    tdelta1inner = np.linalg.norm(v1inner)
    v1outer = outerimagelist[k] - outerimagelist[k - 1]
    tdelta1outer = np.linalg.norm(v1outer)
    tdeltadelta = tdelta1inner - tdelta1outer
    Vvert -= Eholo_vertical[k] * tdeltadelta * kappa1 * 0.5
    return Vvert


def calcEholo_vert(
    image_coords: np.ndarray, tangents: np.ndarray, thresh: float = 1e-3
):
    """Calculate perpendicular tangents for given images and tangents.

    Implements eq. (8) to (13) in [1]. This function seems to crash when all images
    (tangents) lie (point) in a straight line. Currently, this check is done in
    'get_stiff_stress'.
    """
    nimages = len(image_coords)
    tangents_perp = [None for _ in range(nimages)]
    findvertlist = [False for _ in range(nimages)]
    whileN = 0
    while whileN < 1000:
        whileN += 1
        if all(findvertlist):
            break
        for k in range(nimages):
            if findvertlist[k]:
                continue
            tau = tangents[k]
            if k == 0 and findvertlist[1]:
                if findvertlist[1]:
                    tau_nei = tangents_perp[1]
                    a = -np.dot(tau_nei, tau) / np.dot(tau, tau)
                    tangents_perp[k] = a * tau + tau_nei
                    findvertlist[k] = True
            elif k == nimages - 1:
                if findvertlist[-2]:
                    tau_nei = tangents_perp[-2]
                    a = -np.dot(tau_nei, tau) / np.dot(tau, tau)
                    tangents_perp[k] = a * tau + tau_nei
                    findvertlist[k] = True
            else:
                v1 = image_coords[k - 1] - image_coords[k]
                v1 = v1 / np.linalg.norm(v1)
                v1taudot = np.abs(np.dot(v1, tau))
                v2 = image_coords[k + 1] - image_coords[k]
                v2 = v2 / np.linalg.norm(v2)
                v2taudot = np.abs(np.dot(v2, tau))
                if 1.0 - thresh < v1taudot and 1.0 - thresh < v2taudot:
                    if findvertlist[k - 1]:
                        tau_nei = tangents_perp[k - 1]
                        a = -np.dot(tau_nei, tau) / np.dot(tau, tau)
                        tangents_perp[k] = a * tau + tau_nei
                        findvertlist[k] = True
                    elif findvertlist[k + 1]:
                        tau_nei = tangents_perp[k + 1]
                        a = -np.dot(tau_nei, tau) / np.dot(tau, tau)
                        tangents_perp[k] = a * tau + tau_nei
                        findvertlist[k] = True
                elif thresh <= v1taudot:
                    a = -np.dot(v2, tau) / np.dot(v1, tau)
                    tangents_perp[k] = a * v1 + v2
                    tangents_perp[k] /= np.linalg.norm(tangents_perp[k])
                    findvertlist[k] = True
                elif thresh <= v2taudot:
                    a = -np.dot(v1, tau) / np.dot(v2, tau)
                    tangents_perp[k] = a * v2 + v1
                    tangents_perp[k] /= np.linalg.norm(tangents_perp[k])
                    findvertlist[k] = True
                # The elif is from the original code, but it should always evaluted to True.
                # If one of the two conditions wouldn't evalute to True, one of the two
                # conditionals above would run.
                # elif v1taudot < thresh and v2taudot < thresh:
                else:
                    tangents_perp[k] = v1
                    findvertlist[k] = True
                # else:
                # raise Exception("Error in calcEholovertTh")
    for k in range(1, nimages):
        vbefore = tangents_perp[k - 1]
        v = tangents_perp[k]
        if vbefore.dot(v) < 0.0:
            tangents_perp[k] *= -1
    tangents_perp = np.array(tangents_perp)
    return tangents_perp, findvertlist



import numpy as np

from pysisyphus.optimizers.Optimizer import Optimizer

class BacktrackingOptimizer(Optimizer):

    def __init__(self, geometry, alpha, bt_force=5,
                 dont_skip_after=2, bt_max_scale=4,
                 bt_disable=False, **kwargs):
        # Setting some default values
        self.alpha = alpha
        assert(self.alpha > 0), "Alpha must be positive!"
        self.bt_force = bt_force
        self.dont_skip_after = dont_skip_after
        self.bt_max_scale = bt_max_scale
        self.bt_disable = bt_disable
        assert(self.dont_skip_after >= 1)
        self.cycles_since_backtrack = self.bt_force
        self.scale_factor = 0.5

        super(BacktrackingOptimizer, self).__init__(geometry, **kwargs)

        self.alpha0 = self.alpha
        self.alpha_max = self.bt_max_scale * self.alpha0

        # Keep the skipping history to avoid infinite skipping, e.g. always
        # return skip = False if we already skipped in the last n iterations.
        self.skip_log = list()

    def _get_opt_restart_info(self):
        opt_restart_info = {
            "alpha": self.alpha,
            "cycles_since_backtrack": self.cycles_since_backtrack,
        }
        return opt_restart_info

    def _set_opt_restart_info(self, opt_restart_info):
        self.alpha = opt_restart_info["alpha"]
        self.cycles_since_backtrack = opt_restart_info["cycles_since_backtrack"]

    def reset(self):
        if self.alpha > self.alpha0:
            self.alpha = self.alpha0
            self.log(f"Resetting! Current alpha is {self.alpha}. Lowering "
                     f"it to {self.alpha0}.")

    def backtrack(self, cur_forces, prev_forces, reset_hessian=None):
        """Accelerated backtracking line search."""
        if self.bt_disable:
            return False

        epsilon = 1e-3

        rms = lambda f: np.sqrt(np.mean(np.square(f)))
        cur_rms_force = rms(cur_forces)
        prev_rms_force = rms(prev_forces)

        rms_diff = (
            (cur_rms_force - prev_rms_force) /
            np.abs(cur_rms_force + prev_rms_force)
        )

        # Skip tells us if we overshot
        skip = False

        # When the optimiziation is converging cur_forces will
        # be smaller than prev_forces, so rms_diff will be negative
        # and hence smaller than epsilon, which is a positive number.

        # We went uphill, slow alpha
        self.log(f"Backtracking: rms_diff = {rms_diff:.03f}")
        if rms_diff > epsilon:
            self.log(f"Scaling alpha with {self.scale_factor:.03f}")
            # self.alpha = max(self.alpha0*.5, self.alpha*self.scale_factor)
            self.alpha *= self.scale_factor
            skip = True
            self.cycles_since_backtrack = self.bt_force
        # We continnue going downhill, rms_diff is smaller than epsilon
        else:
            self.cycles_since_backtrack -= 1
            # Check if we didn't accelerate in the previous cycles
            if self.cycles_since_backtrack < 0:
                self.cycles_since_backtrack = self.bt_force
                if self.alpha < self.alpha0:
                    # Reset alpha
                    self.alpha = self.alpha0
                    skip = True
                    self.log(f"Reset alpha to alpha0 = {self.alpha0:.4f}")
                else:
                    # Accelerate alpha
                    self.alpha /= self.scale_factor
                    self.log(f"Scaled alpha to {self.alpha:.4f}")

        # Avoid huge alphas
        if self.alpha > self.alpha_max:
            self.alpha = self.alpha_max
            self.log("Didn't accelerate as alpha would become too large. "
                     f"keeping it at {self.alpha}.")

        # Don't skip if we already skipped the previous iterations to
        # avoid infinite skipping.
        if ((len(self.skip_log) >= self.dont_skip_after)
            and all(self.skip_log[-self.dont_skip_after:])):
            self.log(f"already skipped last {self.dont_skip_after} "
                      "iterations don't skip now.")
            skip = False
            if self.alpha > self.alpha0:
                self.alpha = self.alpha0
                self.log("Resetted alpha to alpha0.")
        self.skip_log.append(skip)
        self.log(f"alpha = {self.alpha:.4f}, skip = {skip}")

        return skip



import abc
from collections import namedtuple
import logging

import numpy as np


class LineSearchConverged(Exception):
    def __init__(self, alpha):
        self.alpha = alpha


class LineSearchNotConverged(Exception):
    pass


LineSearchResult = namedtuple(
    "LineSearchResult",
    "converged alpha f_new g_new f_evals df_evals dphi0",
    # defaults=( None, None, None, None,   None,    None),
)


class LineSearch(metaclass=abc.ABCMeta):
    def __init__(
        self,
        p,
        cond="armijo",
        x0=None,
        geometry=None,
        f=None,
        df=None,
        alpha_init=None,
        f0=None,
        g0=None,
        c1=0.1,
        c2=0.9,
        max_cycles=10,
        alpha_min=1e-6,
    ):
        self.p = p
        self.geometry = geometry
        self.f = f
        self.df = df

        geometry_supplied = self.geometry is not None
        assert geometry_supplied or (
            x0 is not None
        ), "Supply either 'geometry' or the starting coordinates 'x0'!"
        assert geometry_supplied or (self.f and self.df), (
            "Supply either 'geometry' with a calculator or the two functions "
            "'f' and 'df' to calculate the energy and its gradient!"
        )

        self.x0 = x0
        if self.geometry:
            self.f = lambda coords: self.geometry.get_energy_at(coords)
            self.df = lambda coords: -self.geometry.get_energy_and_forces_at(coords)[
                "forces"
            ]
            self.x0 = self.geometry.coords.copy()

        self.alpha_init = alpha_init
        self.f0 = f0
        self.g0 = g0
        self.c1 = c1
        self.c2 = c2
        self.max_cycles = max_cycles
        self.alpha_min = alpha_min

        # Store calculated energies & gradients
        self.alpha_fs = {}
        self.alpha_dfs = {}
        self.f_evals = 0
        self.df_evals = 0

        self.dphis = {}

        self.cond_funcs = {
            "armijo": self.sufficiently_decreased,
            "wolfe": self.wolfe_condition,
            "strong_wolfe": self.strong_wolfe_condition,
        }
        self.can_eval_cond_funcs = {
            "armijo": lambda alpha: alpha in self.alpha_fs,
            "wolfe": self.got_alpha_phi_dphi,
            "strong_wolfe": self.got_alpha_phi_dphi,
        }
        self.cond_func = self.cond_funcs[cond]
        self.can_eval_cond_func = self.can_eval_cond_funcs[cond]

        self.logger = logging.getLogger("optimizer")

    def log(self, message):
        self.logger.debug(message)

    def prepare_line_search(self):
        if self.f0 is None:
            self.phi0 = self.get_phi_dphi("f", 0)
        else:
            self.phi0 = self.f0
            self.alpha_fs[0.0] = self.f0
        if self.g0 is None:
            self.dphi0 = self.get_phi_dphi("g", 0)
        else:
            self.dphi0 = self.g0.dot(self.p)
            self.alpha_dfs[0.0] = self.g0

    def check_alpha(self, alpha):
        if (alpha != 0.0) and (np.isnan(alpha) or (alpha < self.alpha_min)):
            raise LineSearchNotConverged()

    def _phi(self, alpha):
        alpha = float(alpha)
        self.check_alpha(alpha)
        try:
            f_alpha = self.alpha_fs[alpha]
        except KeyError:
            self.log(f"\tEvaluating energy for alpha={alpha:.6f}")
            f_alpha = self.f(self.x0 + alpha * self.p)
            self.f_evals += 1
            self.alpha_fs[alpha] = f_alpha
        return f_alpha

    def _dphi(self, alpha):
        alpha = float(alpha)
        self.check_alpha(alpha)
        try:
            df_alpha = self.alpha_dfs[alpha]
            dphi_ = df_alpha.dot(self.p)
        except KeyError:
            self.log(f"\tEvaluating gradient for alpha={alpha:.6f}")
            df_alpha = self.df(self.x0 + alpha * self.p)
            self.df_evals += 1
            self.alpha_dfs[alpha] = df_alpha
            dphi_ = df_alpha.dot(self.p)
            self.dphis[alpha] = dphi_
        return dphi_

    def got_alpha_phi_dphi(self, alpha):
        return (alpha in self.alpha_fs) and (alpha in self.alpha_dfs)

    def get_phi_dphi(self, what, alpha, check=True):
        """Wrapper that handles function/gradient evaluations."""
        alpha = float(alpha)
        whats = "f g fg".split()
        assert what in whats
        calc_funcs = {
            "f": self._phi,
            "g": self._dphi,
        }
        result = [calc_funcs[w](alpha) for w in what]
        # Check if we got both phi and dphi for alpha now. If so we
        # can check if the chosen condition (Wolfe/approx. Wolfe) is
        # satisfied.
        if (
            check
            and (alpha > 0.0)
            and self.can_eval_cond_func(alpha)
            and self.cond_func(alpha)
        ):
            self.log(f"Line search condition is satisfied for α={alpha:.6f}.")
            raise LineSearchConverged(alpha)
        # Dont return a list if only f or g was requested.
        if len(what) == 1:
            result = result[0]
        return result

    def get_fg(self, what, alpha):
        """Lookup raw function/gradient values for a given alpha."""
        whats = "f g fg".split()
        assert what in whats
        lookups = {
            "f": self.alpha_fs,
            "g": self.alpha_dfs,
        }
        result = [lookups[w][alpha] for w in what]
        if len(what) == 1:
            result = result[0]
        return result

    def sufficiently_decreased(self, alpha):
        """Sufficient decrease/Armijo condition."""
        return self._phi(alpha) <= (self.phi0 + self.c1 * alpha * self.dphi0)

    def curvature_condition(self, alpha):
        return self._dphi(alpha) >= self.c2 * self.dphi0

    def strong_curvature_condition(self, alpha):
        return abs(self._dphi(alpha)) <= -self.c2 * self.dphi0

    def wolfe_condition(self, alpha):
        """Normal, not strong, Wolfe condition."""
        return self.sufficiently_decreased(alpha) and self.curvature_condition(alpha)

    def strong_wolfe_condition(self, alpha):
        """Strong wolfe condition."""
        return self.sufficiently_decreased(alpha) and self.strong_curvature_condition(
            alpha
        )

    @abc.abstractmethod
    def run_line_search(self):
        raise NotImplementedError

    def run(self):
        self.prepare_line_search()

        # Normal termination
        try:
            self.log(f"Starting {self.__class__.__name__} line search.")
            alpha = self.run_line_search()
        # Termination in get_phi_dphi
        except LineSearchConverged as lsc:
            alpha = lsc.alpha
        # Failed LineSearch
        except LineSearchNotConverged:
            alpha = None

        result = LineSearchResult(
            converged=bool(alpha),
            alpha=alpha,
            # The gradient at the endpoint of the line search may not
            # have been evaluted, but the function value was always
            # evaluated, except when the line search did not converge.
            f_new=self.alpha_fs.get(alpha, None),
            g_new=self.alpha_dfs.get(alpha, None),
            f_evals=self.f_evals,
            df_evals=self.df_evals,
            dphi0=self.dphi0,
        )
        return result



# pdb2reaction/backends/__init__.py

"""
MLIP backend factory and registry.

Usage::

    from pdb2reaction.backends import create_calculator, create_ase_calculator

    calc = create_calculator(backend="uma", charge=0, spin=1, ...)
    ase_calc = create_ase_calculator(backend="uma", model="uma-s-1p1", ...)
"""

from __future__ import annotations

import importlib
import warnings
from typing import Any, Dict

from .base import BackendError, MLIPCalculator

# Lazy-import registry.  Modules are imported only when the backend is used.
BACKEND_REGISTRY: Dict[str, Dict[str, str]] = {
    "uma": {
        "module": "pdb2reaction.backends.uma",
        "pysis_cls": "UMACalculator",
        "ase_cls": "UMAASECalculator",
    },
    "orb": {
        "module": "pdb2reaction.backends.orb",
        "pysis_cls": "OrbCalculator",
        "ase_cls": "OrbASECalculator",
    },
    "mace": {
        "module": "pdb2reaction.backends.mace",
        "pysis_cls": "MACECalculator",
        "ase_cls": "MACEASECalculator",
    },
    "aimnet2": {
        "module": "pdb2reaction.backends.aimnet2",
        "pysis_cls": "AIMNet2Calculator",
        "ase_cls": "AIMNet2ASECalculator",
    },
}

# Keys accepted by each backend (used to filter calc_cfg kwargs)
_BACKEND_ACCEPTED_KEYS: Dict[str, set] = {
    "uma": {
        "charge", "spin", "device", "freeze_atoms", "hessian_calc_mode",
        "return_partial_hessian", "hessian_double", "print_timing",
        "model", "task_name", "workers", "workers_per_node",
        "max_neigh", "radius", "r_edges", "out_hess_torch", "print_vram",
    },
    "orb": {
        "charge", "spin", "device", "freeze_atoms", "hessian_calc_mode",
        "return_partial_hessian", "hessian_double", "print_timing",
        "model", "precision", "compile_model",
    },
    "mace": {
        "charge", "spin", "device", "freeze_atoms", "hessian_calc_mode",
        "return_partial_hessian", "hessian_double", "print_timing",
        "model", "default_dtype",
    },
    "aimnet2": {
        "charge", "spin", "device", "freeze_atoms", "hessian_calc_mode",
        "return_partial_hessian", "hessian_double", "print_timing",
        "model",
    },
}

# Keys accepted by ASE calculator factories
_ASE_ACCEPTED_KEYS: Dict[str, set] = {
    "uma": {"model", "device", "task_name", "workers", "workers_per_node"},
    "orb": {"model", "device", "precision", "compile_model"},
    "mace": {"model", "device", "default_dtype"},
    "aimnet2": {"model", "device"},
}

VALID_BACKENDS = tuple(BACKEND_REGISTRY.keys())


def _import_cls(backend: str, cls_key: str):
    """Import a class from the backend registry."""
    reg = BACKEND_REGISTRY.get(backend)
    if reg is None:
        raise BackendError(
            f"Unknown backend '{backend}'. Choose from: {', '.join(VALID_BACKENDS)}"
        )
    mod = importlib.import_module(reg["module"])
    return getattr(mod, reg[cls_key])


def _filter_kwargs(kwargs: Dict[str, Any], accepted: set) -> Dict[str, Any]:
    """Return only keys present in *accepted*."""
    return {k: v for k, v in kwargs.items() if k in accepted}


def resolve_backend(backend: str) -> str:
    """Resolve ``'auto'`` to the best available backend (UMA-first)."""
    if backend != "auto":
        return backend
    for key in ("uma", "orb", "mace", "aimnet2"):
        try:
            _import_cls(key, "pysis_cls")
            return key
        except Exception:
            continue
    raise BackendError(
        "No MLIP backend available. Install one of: "
        "fairchem-core (UMA), orb-models (ORB), mace-torch (MACE), aimnet (AIMNet2)."
    )


def create_calculator(backend: str = "uma", **kwargs) -> MLIPCalculator:
    """Create a PySisyphus-compatible MLIP calculator.

    Parameters
    ----------
    backend : str
        One of ``'uma'``, ``'orb'``, ``'mace'``, ``'aimnet2'``, ``'auto'``.
    **kwargs
        Backend-specific and common parameters.  Unknown keys for the
        selected backend are silently ignored.

        Solvent correction keys (``solvent``, ``solvent_model``, ``xtb_cmd``,
        ``xtb_acc``) are extracted and used to wrap the base calculator with
        ``SolventCorrectedCalculator`` when ``solvent != 'none'``.
    """
    # Extract solvent keys before backend filtering
    solvent = str(kwargs.pop("solvent", "none"))
    solvent_model = str(kwargs.pop("solvent_model", "alpb"))
    xtb_cmd = str(kwargs.pop("xtb_cmd", "xtb"))
    xtb_acc = float(kwargs.pop("xtb_acc", 0.2))

    backend = resolve_backend(backend)
    accepted = _BACKEND_ACCEPTED_KEYS.get(backend, set())
    filtered = _filter_kwargs(kwargs, accepted)
    cls = _import_cls(backend, "pysis_cls")
    calc = cls(**filtered)

    # Wrap with solvent correction if enabled
    from .solvent import solvent_correction_enabled
    if solvent_correction_enabled(solvent):
        from .solvent import SolventCorrectedCalculator
        calc = SolventCorrectedCalculator(
            calc,
            solvent=solvent,
            solvent_model=solvent_model,
            xtb_cmd=xtb_cmd,
            xtb_acc=xtb_acc,
        )

    return calc


def create_ase_calculator(backend: str = "uma", **kwargs):
    """Create an ASE-compatible MLIP calculator (used for DMF).

    Parameters
    ----------
    backend : str
        One of ``'uma'``, ``'orb'``, ``'mace'``, ``'aimnet2'``, ``'auto'``.
    **kwargs
        Backend-specific parameters.
    """
    backend = resolve_backend(backend)
    accepted = _ASE_ACCEPTED_KEYS.get(backend, set())
    filtered = _filter_kwargs(kwargs, accepted)
    cls = _import_cls(backend, "ase_cls")
    return cls(**filtered)


__all__ = [
    "BACKEND_REGISTRY",
    "VALID_BACKENDS",
    "BackendError",
    "MLIPCalculator",
    "create_ase_calculator",
    "create_calculator",
    "resolve_backend",
]



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

# Add the project root to the path for autodoc
sys.path.insert(0, os.path.abspath('..'))

# -- Project information -----------------------------------------------------

project = 'pdb2reaction'
copyright = '2025, Takuto Ohmura'
author = 'Takuto Ohmura'

# Get version from package (setuptools_scm)
try:
    from pdb2reaction import __version__
    release = __version__
except ImportError:
    # Fallback for docs build without package installed
    try:
        from pdb2reaction._version import __version__
        release = __version__
    except ImportError:
        release = '0.0.0.dev0'

version = release

# Extract short version (e.g., "0.1.2" from "0.1.2.dev136+gb070dbf49")
short_version = release.split('.dev')[0] if '.dev' in release else release.split('+')[0]

# -- General configuration ---------------------------------------------------

extensions = [
    'myst_parser',                    # Markdown support
    'sphinx.ext.autodoc',             # API documentation from docstrings
    'sphinx.ext.autosummary',         # Generate autodoc summaries
    'sphinx.ext.napoleon',            # Google/NumPy style docstrings
    'sphinx.ext.viewcode',            # Add links to source code
    'sphinx.ext.intersphinx',         # Link to other projects' docs
    'sphinx_copybutton',              # Copy button for code blocks
]

# MyST Parser configuration
myst_enable_extensions = [
    'colon_fence',      # ::: directives
    'deflist',          # Definition lists
    'html_admonition',  # HTML-style admonitions
    'html_image',       # HTML image syntax
    'substitution',     # Substitution syntax
    'tasklist',         # Task lists
    'attrs_inline',     # Inline {#id} attributes
    'attrs_block',      # Block {#id} attributes for headings
]

myst_heading_anchors = 3

# MyST substitutions for version display in Markdown files
# Use {{ version }} or {{ release }} in .md files
myst_substitutions = {
    'version': short_version,
    'release': release,
    'project': project,
}

# Source file suffixes
source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
}

# The master toctree document
master_doc = 'index'

# Patterns to exclude
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'README.md']

# Templates path
templates_path = ['_templates']

# -- Options for HTML output -------------------------------------------------

html_theme = 'furo'

html_theme_options = {
    'light_css_variables': {
        'color-brand-primary': '#2980B9',
        'color-brand-content': '#2980B9',
    },
    'dark_css_variables': {
        'color-brand-primary': '#56B4E9',
        'color-brand-content': '#56B4E9',
    },
    'sidebar_hide_name': False,
    'navigation_with_keys': True,
}

html_title = 'pdb2reaction'
html_short_title = 'pdb2reaction'

# Static files path
html_static_path = ['_static']

# Custom JS — override Furo 3-state toggle to 2-state (light ↔ dark)
html_js_files = ['theme_toggle.js']

# Custom CSS
html_css_files = ['custom.css']

# Favicon (optional)
# html_favicon = '_static/favicon.ico'

# Logo (optional)
# html_logo = '_static/logo.png'

# -- Options for autodoc -----------------------------------------------------

autodoc_default_options = {
    'members': True,
    'member-order': 'bysource',
    'special-members': '__init__',
    'undoc-members': True,
    'exclude-members': '__weakref__',
}

autodoc_typehints = 'description'
autodoc_typehints_format = 'short'

# -- Options for intersphinx -------------------------------------------------

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'torch': ('https://pytorch.org/docs/stable/', None),
}

# -- Options for copy button -------------------------------------------------

copybutton_prompt_text = r'>>> |\.\.\. |\$ |> '
copybutton_prompt_is_regexp = True

# -- Napoleon settings -------------------------------------------------------

napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True



# See [1] 10.1002/jcc.21026

import itertools as it

import numpy as np
import rmsd

from pysisyphus.Geometry import Geometry
from pysisyphus.linalg import get_rot_mat
from pysisyphus.stocastic.Kick import Kick


class FragmentKick(Kick):
    def __init__(
        self,
        geom,
        fragments,
        fix_fragments=None,
        displace_from=None,
        random_displacement=False,
        **kwargs,
    ):
        super().__init__(geom, **kwargs)

        self.fragments = self.get_fragments(fragments)
        if fix_fragments is None:
            fix_fragments = (0,)
        self.fix_fragments = fix_fragments

        if displace_from is None:
            displace_from = list()
        self.displace_from = displace_from
        # True when displace_from is given
        self.random_displacement = self.displace_from or random_displacement

        frag_coords = [self.initial_geom.coords3d[frag] for frag in self.fragments]
        # Shift centroids of the fragment into the cartesian origin
        self.frag_coords = [fc - fc.mean(axis=0) for fc in frag_coords]
        atoms_arr = np.array(self.initial_geom.atoms)
        self.frag_atoms = [atoms_arr[frag] for frag in self.fragments]

        new_coords3d = np.concatenate(self.frag_coords)
        geom = Geometry(list(it.chain(*self.frag_atoms)), new_coords3d.flatten())
        with open("fragments_in_origin.xyz", "w") as handle:
            handle.write(geom.as_xyz())

        self.frag_inds_flat = list(it.chain(*self.fragments))
        # Given an array with fragment coordinates these indices will sort the
        # array into the original atom order.
        self.sort_inds = np.argsort(self.frag_inds_flat)
        self.print_fragments()

    def print_fragments(self):
        lines = self.initial_geom.as_xyz().split("\n")[2:]
        for i, fragment in enumerate(self.fragments):
            self.log(f"Fragment {i}:")
            frag_lines = [lines[j] for j in fragment]
            self.log("\n".join(frag_lines) + "\n")

    def get_fragments(self, fragments):
        fragments = list(fragments)
        # Compare number of atoms defined in fragments with the total
        # number of atoms in the molecule.
        fragment_atoms = list(it.chain(*fragments))
        if len(fragment_atoms) != len(self.atoms):
            self.log(
                "Fragments were not fully specified. " "Determining remaining fragment."
            )
            all_indices = set(range(len(self.atoms)))
            new_fragment = list(all_indices - set(fragment_atoms))
            fragments.append(new_fragment)

        fragments = [np.array(frag) for frag in fragments]
        return fragments

    def get_origin(self):
        if not self.random_displacement:
            return (0, 0, 0)

        # Coordinates of fixed fragment
        frag_coords = self.frag_coords[self.fix_fragments[0]]
        """Here we have two options:
        1.) displace_from is not given, so we use all possible
            atom indices in the fragment.
        2.) displace_from is given, so we restrict the valid
            atom indices. This way some degree of symmetry can
            be considered and the generated structures should
            be more similar. This should reduce the effort in
            matching the geometries later on with the hungarian
            method.

        Consider a stocastic search for minima between benzene and
        a chlorine molecule. As all carbons/hydrogens in the benzene
        are equivalent one could greatly reduce the effort of matching
        atoms later on, when displacing only from one carbon and one
        hydrogen."""
        if not self.displace_from:
            frag_indices = np.arange(frag_coords.shape[0])
        else:
            frag_indices = self.displace_from

        # Select random atom of fixed fragment
        random_ind = np.random.choice(frag_indices, size=1)[0]
        return frag_coords[random_ind]

    def kick_fragment(self, frag_coords):
        R = get_rot_mat()
        kick = self.get_kick()[:3]
        # Fragment rotation
        rot_coords = R.dot(frag_coords.T).T
        # Fragment translation
        rot_kicked_coords = rot_coords + self.get_origin() + kick
        return rot_kicked_coords

    def get_input_geom(self, geom):
        kicked_frags = list()
        for i, fc in enumerate(self.frag_coords):
            if i in self.fix_fragments:
                kicked_frag = fc
            else:
                kicked_frag = self.kick_fragment(fc)
            kicked_frags.append(kicked_frag)
        new_coords3d = np.concatenate(kicked_frags)
        # As the fragments are not necessarily defined in order, especially
        # when we generate the second fragment automatically, we have to
        # sort the coordinates, so they are in the original order. E.g. the
        # fragments ((4, 3), (2, 1, 0)) would result 'new_coords3d' with the
        # coordinates of atom 4 at index 0, coordinates of atom 3 at index 1, etc.
        new_coords3d = new_coords3d[self.sort_inds]
        new_coords = rmsd.kabsch_rotate(new_coords3d, self.initial_coords3d).flatten()
        new_geom = Geometry(self.atoms, new_coords)
        return new_geom



"""Unit tests for pdb2reaction.bond_changes pure helper functions."""

from __future__ import annotations

import pytest


class TestBondStr:
    def test_one_based(self):
        from pdb2reaction.bond_changes import _bond_str

        result = _bond_str(0, 1, ["C", "O"], one_based=True)
        assert result == "C1-O2"

    def test_zero_based(self):
        from pdb2reaction.bond_changes import _bond_str

        result = _bond_str(0, 1, ["C", "O"], one_based=False)
        assert result == "C0-O1"

    def test_same_element(self):
        from pdb2reaction.bond_changes import _bond_str

        result = _bond_str(2, 5, ["H", "H", "C", "N", "O", "C"], one_based=True)
        assert result == "C3-C6"


class TestElementArrays:
    def test_basic(self):
        from pdb2reaction.bond_changes import _element_arrays

        elems, radii = _element_arrays(["C", "O", "N"])
        assert elems == ["C", "O", "N"]
        assert len(radii) == 3
        # Carbon covalent radius should be positive
        assert radii[0] > 0

    def test_lowercase_normalization(self):
        from pdb2reaction.bond_changes import _element_arrays

        elems, radii = _element_arrays(["c", "o"])
        assert elems == ["C", "O"]

    def test_empty(self):
        from pdb2reaction.bond_changes import _element_arrays

        elems, radii = _element_arrays([])
        assert elems == []
        assert len(radii) == 0


class TestResolveDevice:
    def test_cpu(self):
        import torch
        from pdb2reaction.bond_changes import _resolve_device

        dev = _resolve_device("cpu")
        assert dev == torch.device("cpu")

    def test_auto_returns_device(self):
        import torch
        from pdb2reaction.bond_changes import _resolve_device

        dev = _resolve_device("auto")
        assert isinstance(dev, torch.device)



from collections import OrderedDict
import itertools
import logging
from pathlib import Path
import shutil
import subprocess
import tempfile

import h5py
import numpy as np
import pyparsing as pp

from pysisyphus.config import Config
from pysisyphus.helpers_pure import chunks


CIOVL="""mix_aoovl=ao_ovl
a_mo=mos.1
b_mo=mos.2
a_det=dets.1
b_det=dets.2
a_mo_read=2
b_mo_read=2
"""

CIOVL_NO_SAO="""ao_read=-1
same_aos=.true.
a_mo=mos.1
b_mo=mos.2
a_det=dets.1
b_det=dets.2
a_mo_read=2
b_mo_read=2"""


class WFOWrapper2:
    logger = logging.getLogger("wfoverlap")
    matrix_types = OrderedDict((
        ("ovlp", "Overlap matrix"),
        ("renorm", "Renormalized overlap matrix"),
        ("ortho", "Orthonormalized overlap matrix")
    ))

    def __init__(self, overlap_data, calc_number=0, conf_thresh=1e-4, out_dir="./"):
        try:
            self.base_cmd = Config["wfoverlap"]["cmd"]
        except KeyError:
            self.log("WFOverlap cmd not found in ~/.pysisyphusrc!")
        self.overlap_data = overlap_data
        # Should correspond to the attribute of the parent calculator
        self.calc_number = calc_number
        self.conf_thresh = conf_thresh
        self.out_dir = Path(out_dir).resolve()

        with h5py.File(self.overlap_data, "r") as handle:
            self.coords_list = handle["coords"][:]
            self.mo_coeffs_list = handle["mo_coeffs"][:]
            self.ci_coeffs_list = handle["ci_coeffs"][:]

        self.mo_inds_list = list()
        self.from_set_list = list()
        self.to_set_list = list()
        self.turbo_mos_list = list()

        self.name = f"WFOWrapper2_{self.calc_number}"
        self.occ_mo_num, self.virt_mo_num = self.ci_coeffs_list[0,0].shape
        self.mo_num = self.occ_mo_num + self.virt_mo_num
        self.base_det_str = "d"*self.occ_mo_num + "e"*self.virt_mo_num
        self.fmt = "{: .10f}"

        self.iter_counter = 0

        self.set_data()

    @property
    def last_two_coords(self):
        return self.coords_list[-2:]

    def log(self, message):
        self.logger.debug(f"{self.name}, " + message)

    @staticmethod
    def fake_turbo_mos(mo_coeffs):
        """Create a mos file suitable for TURBOMOLE input. All MO eigenvalues
        are set to 0.0. There is also a little deviation in the formatting
        (see turbo_fmt()) but it works ..."""

        def turbo_fmt(num):
            """Not quite the real TURBOMOLE format, but it works ...
            In TURBOMOLE the first character is always 0 for positive doubles
            and - for negative doubles."""
            return f"{num:+20.13E}".replace("E", "D")

        base = "$scfmo    scfconv=7  format(4d20.14)\n# from pysisyphus\n" \
               "{mo_strings}\n$end"

        # WFOverlap expects the string eigenvalue starting at 16, so we have
        mo_str = "{mo_index:>6d}  a      eigenvalue=-.00000000000000D+00   " \
                 "nsaos={nsaos}\n{joined}"
        nsaos = mo_coeffs.shape[0]

        mo_strings = list()
        for mo_index, mo in enumerate(mo_coeffs, 1):
            in_turbo_fmt = [turbo_fmt(c) for c in mo]
            # Combine into chunks of four
            lines = ["".join(chnk) for chnk in chunks(in_turbo_fmt, 4)]
            # Join the lines
            joined = "\n".join(lines)
            mo_strings.append(mo_str.format(mo_index=mo_index, nsaos=nsaos,
                                            joined=joined))
        return base.format(mo_strings="\n".join(mo_strings))

    def ci_coeffs_above_thresh(self, ci_coeffs, thresh=1e-5):
        mo_inds = np.where(np.abs(ci_coeffs) > thresh)
        return mo_inds

    def make_det_string(self, inds):
        """Return spin adapted strings."""
        from_mo, to_mo = inds
        # Until now the first virtual MO (to_mo) has index 0. To subsitute
        # the base_str at the correct index we have to increase all to_mo
        # indices by the number off occupied MO.
        to_mo += self.occ_mo_num
        # Make string for excitation of an alpha electron
        ab = list(self.base_det_str)
        ab[from_mo] = "b"
        ab[to_mo] = "a"
        ab_str = "".join(ab)
        # Make string for excitation of an beta electron
        ba = list(self.base_det_str)
        ba[from_mo] = "a"
        ba[to_mo] = "b"
        ba_str = "".join(ba)
        return ab_str, ba_str

    def generate_all_dets(self, occ_set1, virt_set1, occ_set2, virt_set2):
        """Generate all possible single excitation determinant strings
        from union(occ_mos) to union(virt_mos)."""
        # Unite the respective sets of both calculations
        occ_set = occ_set1 | occ_set2
        virt_set = virt_set1 | virt_set2
        # Genrate all possible excitations (combinations) from the occupied
        # MO set to (and) the virtual MO set.
        all_inds = [(om, vm) for om, vm
                    in itertools.product(occ_set, virt_set)]
        det_strings = [self.make_det_string(inds) for inds in all_inds]
        return all_inds, det_strings

    def make_full_dets_list(self, all_inds, det_strings, ci_coeffs):
        dets_list = list()
        for inds, det_string in zip(all_inds, det_strings):
            ab, ba = det_string
            from_mo, to_mo = inds
            per_state =  ci_coeffs[:,from_mo,to_mo]
            # Drop unimportant configurations, that are configurations
            # having low weights in all states under consideration.
            if np.sum(per_state**2) < self.conf_thresh:
                continue
            # A singlet determinant can be formed in two ways:
            # (up down) (up down) (up down) ...
            # or
            # (down up) (down up) (down up) ...
            # We take this into account by expanding the singlet determinants
            # and using a proper normalization constant.
            # See 10.1063/1.3000012 Eq. (5) and 10.1021/acs.jpclett.7b01479 SI
            per_state *= 1/2**0.5
            as_str = lambda arr: " ".join([self.fmt.format(cic)
                                           for cic in arr])
            ps_str = as_str(per_state)
            mps_str = as_str(-per_state)
            dets_list.append(f"{ab}\t{ps_str}")
            dets_list.append(f"{ba}\t{mps_str}")
        return dets_list

    def set_from_nested_list(self, nested):
        return set([i for i in itertools.chain(*nested)])

    def set_data(self):
        for ci_coeffs in self.ci_coeffs_list:
            mo_inds = [self.ci_coeffs_above_thresh(state)
                       for state in ci_coeffs]
            from_mos, to_mos = zip(*mo_inds)
            from_set = self.set_from_nested_list(from_mos)
            to_set = self.set_from_nested_list(to_mos)
            mo_coeffs = self.mo_coeffs_list[self.iter_counter]
            turbo_mos_fn = f"mos.{self.iter_counter}"
            with open(turbo_mos_fn, "w") as handle:
                handle.write(self.fake_turbo_mos(mo_coeffs))
            self.turbo_mos_list.append(turbo_mos_fn)

            self.mo_inds_list.append(mo_inds)
            self.from_set_list.append(from_set)
            self.to_set_list.append(to_set)
            self.iter_counter += 1

    def get_iteration(self, ind):
        return (self.turbo_mos_list[ind], self.coords_list[ind],
                self.ci_coeffs_list[ind], self.mo_inds_list[ind],
                self.from_set_list[ind], self.to_set_list[ind])

    def make_dets_header(self, cic, dets_list):
        return f"{len(cic)} {self.mo_num} {len(dets_list)}"

    def parse_wfoverlap_out(self, text, type_="ortho"):
        """Returns overlap matrix."""
        header_str = self.matrix_types[type_] + " <PsiA_i|PsiB_j>"
        header = pp.Literal(header_str)
        float_ = pp.Word(pp.nums+"-.")
        psi_bra = pp.Literal("<Psi") + pp.Word(pp.alphas) \
                  + pp.Word(pp.nums) + pp.Literal("|")
        psi_ket = pp.Literal("|Psi") + pp.Word(pp.alphas) \
                  + pp.Word(pp.nums) + pp.Literal(">")
        matrix_line = pp.Suppress(psi_bra) + pp.OneOrMore(float_)

        # I really don't know why this is needed but otherwise I can't parse
        # overlap calculations with the true AO overlap matrix, even though
        # the files appear completely similar regarding printing of the matrices.
        # WTF. WTF!
        text = text.replace("\n", " ")
        parser = pp.SkipTo(header, include=True) \
                 + pp.OneOrMore(psi_ket) \
                 + pp.OneOrMore(matrix_line).setResultsName("overlap")

        result = parser.parseString(text)

        return np.array(list(result["overlap"]), dtype=np.float64)

    def wf_overlap(self, ind1=-2, ind2=-1, ao_ovlp=None):
        iter1 = self.get_iteration(ind1)
        iter2 = self.get_iteration(ind2)

        if ao_ovlp is None:
            mo_coeffs_1 = self.mo_coeffs_list[ind1]
            # mo_coeffs_2 = self.mo_coeffs_list[ind2]
            mo_coeffs_1_inv = np.linalg.inv(mo_coeffs_1)
            ao_ovlp = mo_coeffs_1_inv.dot(mo_coeffs_1_inv.T)

        mos1, coords1, cic1, moi1, fs1, ts1 = iter1
        mos2, coords2, cic2, moi2, fs2, ts2 = iter2
        # Create a fake array for the ground state where all CI coefficients
        # are zero and add it.
        gs_cic = np.zeros_like(cic1[0])
        cic1_with_gs = np.concatenate((gs_cic[None,:,:], cic1))
        cic2_with_gs = np.concatenate((gs_cic[None,:,:], cic2))

        all_inds, det_strings = self.generate_all_dets(fs1, ts1, fs2, ts2)
        # Prepare line for ground state
        gs_coeffs = np.zeros(len(cic1_with_gs))
        # Ground state is 100% HF configuration
        gs_coeffs[0] = 1
        gs_coeffs_str = " ".join([self.fmt.format(c)
                                  for c in gs_coeffs])
        gs_line = f"{self.base_det_str}\t{gs_coeffs_str}"
        dets1 = [gs_line] + self.make_full_dets_list(all_inds,
                                                     det_strings,
                                                     cic1_with_gs)
        dets2 = [gs_line] + self.make_full_dets_list(all_inds,
                                                     det_strings,
                                                     cic2_with_gs)
        header1 = self.make_dets_header(cic1_with_gs, dets1)
        header2 = self.make_dets_header(cic2_with_gs, dets2)

        backup_path = self.out_dir / f"wfo_{self.calc_number}.{ind1:03d}_{ind2:03d}"
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_path = Path(tmp_dir)
            self.log(f"Calculation in {tmp_dir}")
            shutil.copy(mos1, tmp_path / "mos.1")
            shutil.copy(mos2, tmp_path / "mos.2")
            dets1_path = tmp_path / "dets.1"
            with open(dets1_path, "w") as handle:
                handle.write(header1+"\n"+"\n".join(dets1))
            dets2_path = tmp_path / "dets.2"
            with open(dets2_path, "w") as handle:
                handle.write(header2+"\n"+"\n".join(dets2))

            # Decide wether to use a double molecule overlap matrix or
            # (approximately) reconstruct the ao_ovlp matrix from the MO
            # coefficients.
            if ao_ovlp is None:
                ciovl_in = CIOVL_NO_SAO
                self.log("Got no ao_ovl-matrix. Using ao_read=-1 and "
                         "same_aos=.true. to reconstruct the AO-overlap matrix!")
            else:
                ciovl_in = CIOVL
                ao_header = "{} {}".format(*ao_ovlp.shape)
                ao_ovl_path = tmp_path / "ao_ovl"
                np.savetxt(ao_ovl_path, ao_ovlp, fmt="%22.15E", header=ao_header,
                           comments="")

            ciovl_fn = "ciovl.in"
            with open(tmp_path / ciovl_fn, "w") as handle:
                handle.write(ciovl_in)

            # Create a backup of the whole temporary directory
            try:
                shutil.rmtree(backup_path)
            except FileNotFoundError:
                pass
            shutil.copytree(tmp_dir, backup_path)

            cmd = f"{self.base_cmd} -m 4000 -f {ciovl_fn}".split()
            result = subprocess.Popen(cmd, cwd=tmp_path,
                                      stdout=subprocess.PIPE)
            result.wait()
            stdout = result.stdout.read().decode("utf-8")
        if "differs significantly" in stdout:
            self.log("WARNING: Orthogonalized matrix differs significantly "
                     "from original matrix! There is probably mixing with "
		     "external states.")

        wfo_log_fn = self.out_dir / f"wfo_{self.calc_number}.{ind1:03d}_{ind2:03d}.out"
        with open(wfo_log_fn, "w") as handle:
            handle.write(stdout)
        # Also copy the WFO-output to the input backup
        shutil.copy(wfo_log_fn, backup_path)

        matrices = [self.parse_wfoverlap_out(stdout, type_=key)
                    for key in self.matrix_types.keys()]

        reshaped_mats = [mat.reshape(-1, len(cic2_with_gs))
                         for mat in matrices]
        for key, mat in zip(self.matrix_types.keys(), reshaped_mats):
            mat_fn = backup_path / f"{key}_mat.dat"
            np.savetxt(mat_fn, mat)

        # for mat in reshaped_mats:
            # print(mat)
        return reshaped_mats

    def all_overlaps(self):
        for i in range(self.iter_counter-1):
            ind1 = i
            ind2 = i+1
            print(f"Doing overlaps between {ind1} and {ind2}")
            overlap_mats = self.wf_overlap(ind1, ind2)
            ovlp_mat_fn = f"wf_ovlp_mat_{ind1:03d}_{ind2:03d}.dat"
            np.savetxt(ovlp_mat_fn, overlap_mats[-1])
            print(f"Wrote '{ovlp_mat_fn}'")

    def __str__(self):
        return self.name



# [1] https://doi.org/10.1002/jcc.26495
#     Habershon, 2021

import itertools as it

import numpy as np

from pysisyphus.elem_data import COVALENT_RADII as CR


class AtomAtomTransTorque:
    def __init__(
        self,
        geom,
        frags,
        A_mats,
        kappa=2.0,
    ):
        """Atom-atom translational and torque forces.

        See A.5. [1], Eq. (A6).
        """
        self.geom = geom
        self.frags = frags
        self.A_mats = A_mats
        self.kappa = kappa

        self.frag_num = len(self.frags)
        self.pair_inds = list(it.permutations(range(self.frag_num), 2))
        self.frag_sizes_sq = [len(self.frags[m]) ** 2 for m, _ in self.pair_inds]

        frag_atoms = [[self.geom.atoms[a].lower() for a in frag] for frag in self.frags]
        frag_cov_rads = [np.array([CR[fa.lower()] for fa in fas]) for fas in frag_atoms]
        self.avg_cov_radii = list()
        for m, n in self.pair_inds:
            m_cov_rads = frag_cov_rads[m]
            n_cov_rads = frag_cov_rads[n]
            mn_avg_cov_radii = (m_cov_rads[:, None] + n_cov_rads[None, :]) / 2
            self.avg_cov_radii.append(mn_avg_cov_radii)

        def phi_func(A, a):
            """See (A6) in [1]."""
            return 1.5 if a in A else 2.0

        self.phis = list()
        for m, n in self.pair_inds:
            key = (m, n)
            A = self.A_mats[key]
            self.phis.append(np.array([phi_func(A, a) for a in self.frags[m]]))

    # def get_forces_naive(self, atoms, coords):
        # def p(s):
            # print(f"REF: {s}")

        # c3d = coords.reshape(-1, 3)
        # gs = [c3d[frag].mean(axis=0) for frag in self.frags]

        # zt = np.zeros((self.frag_num, 3))
        # zr = np.zeros((self.frag_num, 3))
        # for (m, n), phis in zip(self.pair_inds, self.phis):
            # p(f"m={m}, n={n}")
            # mfrag = self.frags[m]
            # nfrag = self.frags[n]
            # A = self.A_mats[(m, n)]
            # g = gs[m]
            # N = 0
            # for a in mfrag:
                # ra = c3d[a]
                # ramg = ra - g
                # for b in nfrag:
                    # rb = c3d[b]
                    # rmr = rb - ra
                    # a_atm = atoms[a].lower()
                    # b_atm = atoms[b].lower()
                    # cab = (CR[a_atm] + CR[b_atm]) / 2
                    # phi = 1.5 if a in A else 2
                    # nrmr = np.linalg.norm(rmr)
                    # H = 1 if nrmr < (phi * cab) else 0
                    # y = cab * rmr / np.linalg.norm(rmr) - rmr
                    # p(f"y_(a={a},b={b})={y}")
                    # ny = np.linalg.norm(y)
                    # zt[m] += H * y.dot(ramg) * y / ny
                    # p(f"dot={y.dot(ramg)}")
                    # zr[m] += H * np.cross(y, ramg)
                    # p(f"zt={zt}")
                    # p(f"zr={zr}")
                    # N += H
            # N *= self.frag_sizes_sq[m]
            # p(f"N={N}")
            # if N == 0:
                # N_inv = 0
            # else:
                # N_inv = 1 / N
            # zt[m] *= N_inv
            # zr[m] *= N_inv

        # forces = np.zeros_like(c3d)
        # for m, mfrag in enumerate(self.frags):
            # forces[mfrag] = np.cross(-zr[m], c3d[mfrag] - gs[m]) + zt[m]
        # forces *= self.kappa
        # # return zt, zr
        # print("REF")
        # print(forces)
        # return {"energy": 1, "forces": forces.flatten()}

    def get_forces(self, atoms, coords):
        c3d = coords.reshape(-1, 3)

        if len(self.pair_inds) == 0:
            return {"energy": 1, "forces": np.zeros_like(coords)}

        frag_coords = [c3d[frag] for frag in self.frags]
        frag_centroids = np.array([c3d[frag].mean(axis=0) for frag in self.frags])
        frag_centered = [
            frag_coords[m] - frag_centroids[m] for m in range(self.frag_num)
        ]
        Hs = np.zeros(self.frag_num)

        zt = np.zeros((self.frag_num, 3))
        zr = np.zeros((self.frag_num, 3))
        for (m, n), phi, avg_cov_radii in zip(
            self.pair_inds, self.phis, self.avg_cov_radii
        ):
            mfrag = self.frags[m]
            mcoords3d = c3d[mfrag]
            ncoords3d = c3d[self.frags[n]]
            coord_diffs = ncoords3d[None, :] - mcoords3d[:, None]
            norms = np.linalg.norm(coord_diffs, axis=2)
            H = norms < phi[:, None] * avg_cov_radii
            H_sum = H.sum()
            if H_sum == 0:
                continue
            Hs[m] += H_sum
            y = (avg_cov_radii / norms - 1)[:, :, None] * coord_diffs
            ynorms = np.linalg.norm(y, axis=2)
            mcoords3d_centered = frag_centered[m]
            dot = np.abs(np.einsum("ijk,ik->ij", y, mcoords3d_centered))
            zt[m] += (H[:, :, None] * dot[:, :, None] * y / ynorms[:, :, None]).sum(
                axis=(0, 1)
            )
            zr[m] += (H[:, :, None] * np.cross(y, mcoords3d_centered[:, None, :])).sum(
                axis=(0, 1)
            )
        N = self.frag_sizes_sq * Hs
        N_invs = np.divide(1, N, out=np.zeros_like(N), where=N != 0)
        zt *= N_invs[:, None]
        zr *= N_invs[:, None]

        forces = np.zeros_like(c3d)
        for m, mfrag in enumerate(self.frags):
            forces[mfrag] = np.cross(-zr[m], frag_centered[m]) + zt[m]
        forces *= self.kappa

        return {"energy": 1, "forces": forces.flatten()}



#!/usr/bin/env python

# [1] https://doi.org/10.1002/qua.560540202
#     Transformation between Cartesian and pure spherical harmonic Gaussians
#     Schlegel, Frisch, 1995

import argparse
import functools
from cmath import sqrt
from math import floor
import sys
from typing import Dict, List, Tuple

import numpy as np
from numpy.typing import NDArray
from scipy.special import factorial as fact
from scipy.special import binom

from pysisyphus.wavefunction.helpers import canonical_order
from pysisyphus.wavefunction.normalization import get_lmn_factors


ZERO_THRESH = 1e-14


@functools.cache
def cart2sph_coeff(lx: int, ly: int, lz: int, m: int) -> complex:
    """Coefficient to convert Cartesian to spherical harmonics.

    Eq. (15) in [1].

    Paramters
    ---------
    lx
        Cartesian quantum number.
    ly
        Cartesian quantum number.
    lz
        Cartesian quantum number.
    m
        Magnetic quantum number.

    Returns
    -------
    coeff
        Contribution of the given Cartesian basis function to the spherical
        harmonic Y^(lx + ly + lz)_m.
    """

    l = lx + ly + lz
    assert -l <= m <= l

    j = (lx + ly - abs(m)) / 2
    if (j % 1) != 0:  # Return when j is not integer
        return 0
    j = int(j)

    lfact = fact(l)  # l!
    lmm = l - abs(m)  # (l - |m|)
    sign = 1 if m >= 0 else -1

    coeff = 0.0j
    for i in range(floor(lmm / 2) + 1):
        i_fact = (
            binom(l, i)
            * binom(i, j)
            * (-1) ** i
            * fact(2 * l - 2 * i)
            / fact(lmm - 2 * i)
        )
        k_sum = 0.0
        for k in range(j + 1):
            prefact = (-1) ** (sign * (abs(m) - lx + 2 * k) / 2)
            k_sum += prefact * binom(j, k) * binom(abs(m), lx - 2 * k)
        coeff += i_fact * k_sum
    coeff *= sqrt(
        fact(2 * lx)
        * fact(2 * ly)
        * fact(2 * lz)
        * lfact
        * fact(lmm)
        / (fact(2 * l) * fact(lx) * fact(ly) * fact(lz) * fact(l + abs(m)))
    )
    coeff *= 1 / (2**l * lfact)

    return coeff


@functools.cache
def cart2sph_coeffs_for(
    l: int,
    real: bool = True,
    zero_small: bool = True,
    zero_thresh: float = ZERO_THRESH,
) -> NDArray:
    all_coeffs = list()
    for lx, ly, lz in canonical_order(l):
        coeffs = list()
        for m in range(-l, l + 1):
            coeff = cart2sph_coeff(lx, ly, lz, m)
            # Form linear combination for m != 0 to get real spherical orbitals.
            if real and m != 0:
                coeff_minus = cart2sph_coeff(lx, ly, lz, -m)
                sign = 1 if m < 0 else -1
                coeff = (coeff + sign * coeff_minus) / sqrt(sign * 2)
            coeffs.append(coeff)
        all_coeffs.append(coeffs)
    C = np.array(all_coeffs, dtype=complex)
    if real:
        assert real == ~np.iscomplex(C).all()  # C should be real
        C = np.real(C)
    if zero_small:
        mask = np.abs(C) <= zero_thresh
        C[mask] = 0.0
    # Final shape will be (spherical, Cartesian)
    # Return read-only view; otherwise it would be too easy to mess up the
    # cached result by inplace operations on C.
    C = C.T.view()
    C.flags.writeable = False
    return C


@functools.cache
def expand_sph_quantum_numbers(
    Lm, zero_thresh=ZERO_THRESH, with_lmn_factors=False
) -> Tuple[NDArray[float], List[Tuple[int, int, int]]]:
    """Factors and Cart. angular momentum vectors for given sph. quantum numbers."""
    L, m = Lm
    m += L  # Map m from [-L, L] onto [0, 2*L]
    C = cart2sph_coeffs_for(L)  # shape (spherical, Cartesian)
    # Include lmn-factor that appears when contracted functions are normalized
    # according to 'normalization.norm_cgto_lmn(L).
    if with_lmn_factors:
        lmn_factors = get_lmn_factors(L)
        C = C * lmn_factors[None, :]
    indices, *_ = np.where(np.abs(C[m]) > zero_thresh)
    factors = C[m, indices].copy()

    cart_lmn = [lmn for i, lmn in enumerate(canonical_order(L)) if i in indices]
    return factors, cart_lmn


def cart2sph_coeffs(
    l_max: int,
    **kwargs,
) -> Dict[int, NDArray]:
    coeffs = {l: cart2sph_coeffs_for(l, **kwargs) for l in range(l_max + 1)}
    return coeffs


def cart2sph_nlms(l_max: int) -> Dict[int, Tuple[Tuple[int, int, int]]]:
    nlms = dict()
    for l in range(l_max + 1):
        n = l
        nlms[l] = tuple([(n, l, m) for m in range(-l, l + 1)])
    return nlms


def print_coeffs(l: int, C: NDArray):
    print("".join([f"{m: >13d}" for m in range(-l, l + 1)]))
    for (lx, ly, lz), line in zip(canonical_order(l), C):
        fmt = "{:>12.4f}" * len(line)
        verb = "".join(["x"] * lx + ["y"] * ly + ["z"] * lz)
        print(f"{verb}: {fmt.format(*line)}")


def parse_args(args):
    parser = argparse.ArgumentParser()

    parser.add_argument("l_max", type=int, default=3)
    parser.add_argument("--not-normalized", action="store_true")

    return parser.parse_args(args)


def run():
    args = parse_args(sys.argv[1:])

    l_max = args.l_max
    normalized = not args.not_normalized

    Cs = cart2sph_coeffs(l_max, normalized=normalized)
    for l, C in Cs.items():
        print(f"{l=}")
        print_coeffs(l, C.T)
        print()


if __name__ == "__main__":
    run()



from pathlib import Path
import tarfile

from fabric import Connection
import yaml

from pysisyphus.calculators.Calculator import Calculator
from pysisyphus.helpers_pure import json_to_results


class Remote(Calculator):
    def __init__(self, remote_calc, host, prefix="", **kwargs):
        super().__init__(**kwargs)

        self.remote_calc = remote_calc
        self.host = host
        self.prefix = prefix

        self.calc_inp = yaml.dump(self.remote_calc)
        self.tar_fn = "pysis.tar.gz"
        self.yaml_fn = "inp.yaml"
        self.run_dict = {
            "geom": {
                "fn": None,
            },
            "calc": {
                "run_func": None,
            },
        }
        self.run_dict["calc"].update(remote_calc)

    def run_calculation(self, atoms, coords, run_func="get_energy"):
        con = Connection(self.host)
        res = con.run("mktemp -d", hide=True)
        tmp_dir = res.stdout.strip()

        xyz_str = self.prepare_xyz_string(atoms, coords)
        run_dict = self.run_dict.copy()
        run_dict["geom"]["fn"] = xyz_str
        # Inline xyz coordinates
        run_dict["calc"]["run_func"] = run_func

        # Create YAML input for remote calculation
        with open(self.yaml_fn, "w") as handle:
            yaml.dump(run_dict, handle)

        with con.cd(tmp_dir):
            tmp_parent = Path(tmp_dir).parent
            tar_target = tmp_parent / self.tar_fn
            con.put(self.yaml_fn, tmp_dir)
            # yaml_str = yaml.dump(run_dict)
            # con.run(f"cat > {self.yaml_fn} <<EOL\n{yaml_str}\nEOL")
            # Execute pysis with prefix, e.g. activation of conda env or venv
            with con.prefix(self.prefix):
                con.run(f"pysis {self.yaml_fn}", hide=True)
            # Pack the whole directory
            con.run(f"tar -czf {tar_target} .")
            # and download it.
            con.get(tar_target, self.tar_fn)
        con.run(f"rm -r {tmp_dir}")
        return self.parse_results()

    def parse_results(self):
        with tarfile.open(self.tar_fn, "r:gz") as tfile:
            as_json = tfile.extractfile("./calculator_000.000.results").read()
            results = json_to_results(as_json)
        return results

    def get_energy(self, atoms, coords, **prepare_kwargs):
        return self.run_calculation(atoms, coords, run_func="get_energy")

    def get_forces(self, atoms, coords, **prepare_kwargs):
        return self.run_calculation(atoms, coords, run_func="get_forces")

    def get_hessian(self, atoms, coords, **prepare_kwargs):
        return self.run_calculation(atoms, coords, run_func="get_hessian")



import numpy as np


class DummyColvar:
    def eval(self, x):
        assert np.isscalar(
            x
        ), f"DummyColvar can only be used for scalar inputs, but got {x}!"
        return x, 1


class Gaussian:
    def __init__(self, w=1, s=1, x0=None, colvar=None, dump_name=None):
        """
        See:
            https://doi.org/10.1016/j.cpc.2018.02.017

        V(f(x)) = w * exp(-(f(x) - f0)**2 / (2*s**2))

        F(x) = -dV/dx = -dV/df * df/dx

        See also:
            https://doi.org/10.1103/PhysRevLett.90.238302
        """
        # Gaussian height w and standard deviation s are read-only properties.
        # This way they can't be accidentally altered, which would invalidate the
        # values precomputed below.
        self._w = w
        self._s = s
        if x0 is None:
            x0 = np.array(())
        self.x0 = np.ravel(x0)

        # Collective variable
        if colvar is None:
            colvar = DummyColvar()
        self.colvar = colvar
        self.dump_name = dump_name
        if self.dump_name:
            self.dump_fn = (
                dump_name
                if self.dump_name.endswith(".gau")
                else f"{self.dump_name}.gau"
            )
        else:
            self.dump_fn = None

        # Reset file and write header
        if self.dump_fn:
            with open(self.dump_fn, "w") as handle:
                handle.write(f"# {dump_name}\n# step s w center\n")

        # Store some values, to avoid recalculation
        self.s2 = self.s ** 2
        self.one_over_2s2 = 1 / (2 * self.s2)
        self.minus_w_over_s2 = -self.w / self.s2

    @property
    def w(self):
        return self._w

    @property
    def s(self):
        return self._s

    def calculate(self, coords, x0=None, gradient=False):
        """Return potential and gradient for Gaussian(s) centered at x0."""

        if x0 is None:
            x0 = self.x0
        x, cr_grad = self.colvar.eval(coords)
        # TODO: allow colvar to supply a callback, that overrides the difference
        # below. This will be needed to support periodic colvars, like torsion.
        # There we can't just calculate the naive difference.
        diff = x - np.atleast_1d(x0)
        exp_ = np.exp(-(diff ** 2) * self.one_over_2s2)
        to_return = self.w * exp_.sum()

        if gradient:
            try:
                grad = np.einsum("i,i,ijk->jk", diff, exp_, cr_grad[None, :, :])
            except TypeError:
                grad = diff * exp_ * cr_grad

            # Finalize & append gradient
            grad *= self.minus_w_over_s2
            to_return = to_return, grad
        return to_return

    def value(self, coords, x0=None):
        return self.calculate(coords, x0)

    def gradient(self, coords, x0=None):
        _, grad = self.calculate(coords, x0, gradient=True)
        return grad

    def eval(self, coords, x0=None):
        return self.calculate(coords, x0, gradient=True)

    def dump(self, step, s, w, center):
        line = f"{step:08d} {s:.12f} {w:.12f} {center:.12f}\n"
        with open(self.dump_fn, "a") as handle:
            handle.write(line)



import numpy as np

from pysisyphus.irc.IRC import IRC
from pysisyphus.TableFormatter import TableFormatter


class Euler(IRC):

    def __init__(self, geometry, step_length=0.01, **kwargs):
        super(Euler, self).__init__(geometry, step_length, **kwargs)

        step_header = "E/au ΔE(TS)/au |gradient|".split()
        step_fmts = [".6f", ".6f", ".4f"]
        self.step_formatter = TableFormatter(step_header, step_fmts, 10)

    def step(self):
        grad = self.mw_gradient
        grad_norm = np.linalg.norm(grad)

        # Step downhill, against the gradient
        step_direction = -grad / grad_norm
        self.mw_coords += self.step_length*step_direction
